#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.
'''
A simple traffic utility which starting standalone TCP/UDP client/servers.

USAGE:
---------
# Start TCP Server on port 5649
$ python -mlydian.utils.traffic -t -v -s -p localhost 5649

# Start TCP Client to send ping to above server.
$ python -mlydian.utils.traffic -t -v -c -p localhost 5649
'''

import argparse
import signal
import sys

from lydian.traffic import client as hclient
from lydian.traffic import server as hserver
from lydian.utils.common import is_py3


def _print(msg):
    if is_py3():
        print(msg)


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        '-s', '--server', action="store_true",
        help="Run server")
    parser.add_argument(
        '-c', '--client', action="store_true",
        help="Run client")
    parser.add_argument(
        '-i', '--ipv6', action="store_true",
        help="Use IPv6 client / servers.")
    parser.add_argument(
        '-p', '--port', nargs='+',
        help="Port to connect")
    parser.add_argument(
        '-t', '--tcp', action="store_true",
        help="Run TCP traffic.")
    parser.add_argument(
        '-u', '--udp', action="store_true",
        help="Run UDP traffic.")
    parser.add_argument(
        '-v', '--verbose', action="store_true",
        help="Verbose mode")

    args = parser.parse_args()

    verbose = bool(args.verbose)

    Client, Server = None, None

    if args.tcp:
        Client, Server = hclient.TCPClient, hserver.TCPServer
    elif args.udp:
        Client, Server = hclient.UDPClient, hserver.UDPServer

    assert(args.port)
    try:
        host = args.port[0]
        port = int(args.port[1])
    except Exception:
        _print("Invalid host/port")
        raise

    def ping_handler(payload, data):
        if is_py3:
            payload = payload.decode()
            data = data.decode()
        if payload == data:
            _print("Success : Sent: %s , received: %s" % (payload, data))
        else:
            _print("Failure : Sent: %s , received: %s" % (payload, data))

    ipv6 = args.ipv6

    _client, _server = None, None
    def signal_handler(sig, frame):
        if _client:
            _client.close()
            print ("\nClosed the client connection")
        if _server:
            _server.close()
            print ("\nClosed the server connection")
        sys.exit(0)

    signal.signal(signal.SIGINT, signal_handler)

    if args.server:
        ipv6 = args.ipv6
        _server = Server(port=port, verbose=verbose, ipv6=ipv6)
        _server.start()
    else:
        _client = Client(server=host, port=port, verbose=verbose,
                         handler=ping_handler)
        _client.start(tries=10)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.

import logging
import os
import sqlite3
import unittest
import uuid

from lydian.apps.rules import RulesApp


DB_NAME = 'rules_test.db'
log = logging.getLogger(__name__)


class RulesAppTest(unittest.TestCase):

    DUMMY_RULE = {
        'ruleid': '%s' % uuid.uuid4(),
        'src': '127.0.0.1',
        'dst': '127.0.0.1',
        'protocol': 'TCP',
        'port': 9465,
        'connected': False
    }

    def setUp(self):
        if os.path.exists(DB_NAME):
            os.remove(DB_NAME)

    def _get_new_app(self):
        class DummyTrafficManager(object):

            def start(self, ruleid):
                log.info("Starting Traffic for rule %s ", ruleid)

            def stop(self, ruleid):
                pass

        return RulesApp(DummyTrafficManager(), DB_NAME)

    def test_add(self):
        self.rulesApp = self._get_new_app()

        # TEST : Add one rule
        self.rulesApp.add(self.DUMMY_RULE)

        # TEST : Adding same rule would lead to a warning
        # but no assertion / exception to break the run.
        self.rulesApp.add(self.DUMMY_RULE)

        rules = []
        for x in range(5):
            rule = {k: v for k, v in self.DUMMY_RULE.items()}
            rule['ruleid'] = '%s' % uuid.uuid4()
            rules.append(rule)

        # TEST : Adding multiple rules.
        self.rulesApp.add_rules(rules)

        self.rulesApp.save_to_db()
        self.rulesApp.close()

        # Get New instance of rulesApp()
        # If persistence works, all the rules must be brought up fine.
        self.rulesApp = self._get_new_app()

        self.rulesApp.load_from_db()

        # TEST : All the rules must be prsent.
        rules.append(self.DUMMY_RULE)

        for rule in rules:
            assert self.rulesApp.get(rule['ruleid'])

        # TEST : Diable / Enable of rules should work.
        self.rulesApp.disable(rules[0]['ruleid'])
        # disable and enable the rule
        self.rulesApp.disable(rules[1]['ruleid'])
        self.rulesApp.enable(rules[1]['ruleid'])

        assert not self.rulesApp.is_enabled(rules[0]['ruleid'])
        assert self.rulesApp.is_enabled(rules[1]['ruleid'])

    def tearDown(self):
        os.remove(DB_NAME)



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.
'''
Common constants for Lydian application.
'''
import os
import platform


LYDIAN_PORT = 5649

NAMESPACE_INTERFACE_NAME_PREFIXES = ["veth", "eth"]

#### CONSTANTS ####
SYSTEM = platform.system()
LINUX_OS = True if SYSTEM == 'Linux' else False
MAC_OS = True if SYSTEM == 'Darwin' else False
WIN_OS = True if SYSTEM == 'Windows' else False

# Logging Constants
LINUX_LOG_DIR = os.environ.get('LINUX_LOG_DIR', "/var/log/lydian")
WIN_LOG_DIR = os.environ.get('WIN_LOG_DIR', "C:\\lydian\\log")
if LINUX_OS or MAC_OS:
    LOG_DIR = LINUX_LOG_DIR
elif WIN_OS:
    LOG_DIR = WIN_LOG_DIR
LOG_FILE = "lydian.log"



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.
"""
-App for Resource Monitoring.
-Collects System CPU/Memory as well as AXON CPU/ Memory.
"""
import logging
import os
import queue
import threading
import time

import psutil

from lydian.apps.base import BaseApp, exposify
from lydian.recorder.record import ResourceRecord

log = logging.getLogger(__name__)


@exposify
class ResourceMonitor(BaseApp):

    REPORT_INTERVAL = 2
    def __init__(self, rqueue, interval=None, proc_name='runner'):
        """
        A simple resource monitor that writes cpu / memory percentage
        to wavefront at requested interval.
        """
        self._rqueue = rqueue    # records queue to put records onto.
        self._interval = interval or self.REPORT_INTERVAL
        self._switch = threading.Event()
        self._proc_name = proc_name
        self._thread = None

    def _run(self):
        p = psutil.Process(os.getpid())
        while self._switch.is_set():
            sys_cpu_percent = round(psutil.cpu_percent(), 2)
            sys_mem_percent = round(psutil.virtual_memory().percent, 2)
            sys_net_conns = int(len(psutil.net_connections()))

            lydian_cpu_percent = round(p.cpu_percent(), 2)
            lydian_mem_percent = round(p.memory_percent(), 2)
            lydian_net_conns = int(len(p.connections()))

            rec = ResourceRecord(sys_cpu_percent, sys_mem_percent,
                                 sys_net_conns,
                                 lydian_cpu_percent, lydian_mem_percent,
                                 lydian_net_conns)
            try:
                self._rqueue.put(rec, block=False, timeout=2)
            except queue.Full:
                log.error("Cann't put Resource record %r into the queue.",
                          rec)

            time.sleep(self._interval)

    def is_running(self):
        """
        Returns True if Rescoures are being monitored else False.
        """
        return self._thread and self._thread.is_alive()

    def stop(self):
        """
        Stops Resource Monitoring.
        """
        self._switch.clear()
        if self.is_running():
            self._thread.join()
            self._thread = None
        log.info("Stopped resource monitoring.")

    def start(self):
        """
        Starts Resource monitoring (in a separate thread)
        """
        self._switch.set()
        if not self._thread:
            self._thread = threading.Thread(target=self._run)
            self._thread.setDaemon(True)
            self._thread.start()
        log.info("Started resource monitoring.")



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.

import logging

import lydian.traffic.core as core
import lydian.traffic.task as task

log = logging.getLogger(__name__)


class TrafficManager(object):
    TASK_TYPE = None

    def __init__(self):
        self._traffic_tasks = {}

    def key(self, trule):
        raise NotImplementedError("'key' method must be implemented.")

    def _create_task(self, trule):
        raise NotImplementedError("'key' method must be implemented.")

    def add_task(self, trule):
        """
        Relationship { Traffic Client : Traffic Rule } is 1:1.
        """
        key = self.key(trule)
        assert key, "Invalid ruleid"
        if key in self._traffic_tasks:
            log.error("Traffic task (%s) already running for ruleid: %s",
                      self.TASK_TYPE, trule.ruleid)
            return

        task = self._create_task(trule)
        if trule.enabled:
            task.start()    # Start the task if enabled.
        self._traffic_tasks[key] = task

    def start(self, trule):
        """ Start a Traffic task (again). """
        key = self.key(trule)
        if key in self._traffic_tasks:
            self._traffic_tasks[key].start()

    def stop(self, trule):
        """ Stop a Traffic task. """
        key = self.key(trule)
        if key in self._traffic_tasks:
            self._traffic_tasks[key].stop()

    def close(self):
        for _, task in self._traffic_tasks.items():
            task.close()
        self._traffic_tasks = {}

    def num_tasks(self):
        return len(self._traffic_tasks)


class ClientManager(TrafficManager):
    """
    Manages Traffic Client on a host/vm.
    """
    TASK_TYPE = task.TrafficTask.CLIENT

    def __init__(self, record_queue):
        self._record_queue = record_queue
        super(ClientManager, self).__init__()

    def key(self, trule):
        # A client can be uniquely identified with the rule it is
        # attached to. Two clients can be alike except for the
        # ruleid they are attached to.
        return trule.ruleid

    def _create_task(self, trule):
        return task.TrafficClientTask(self._record_queue, trule)


class ServerManager(TrafficManager):
    """
    Manages Traffic Client on a host/vm.
    Relationship { Traffic Server : Traffic Rule } is 1:N. Basically,
    there can be multiple Traffic Rules requiring to start same server
    and that would be a valid request.
    """
    TASK_TYPE = task.TrafficTask.SERVER

    def key(self, trule):
        # A server can be uniquely identified by its target host,
        # protocol and port associated.
        return (trule.dst_target.name, trule.protocol, trule.port)

    def _create_task(self, trule):
        return task.TrafficServerTask(trule)



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.

import logging
import threading


class Connection(object):

    # Default server parameters
    MAX_CONNS = 20
    DEFAULT_TCP_SERVER_PORT = 5649
    DEFAULT_UDP_SERVER_PORT = 5648

    # Default client parameters
    TCP_CLIENT_PAYLOAD = "Dunkirk!!"

    # We are writing only a connection validation utility.
    # For our purpose, 50 bytes is good enough as payload
    # data.
    MAX_PAYLOAD_SIZE = 4096

    def __init__(self, verbose=False):
        self.log = logging.getLogger(__name__)
        self.verbose = verbose
        self._stop_event = threading.Event()
        self.socket = None

    def _create_socket(self):
        raise NotImplementedError("%s::_create_socket not implemented." % type(self).name)

    def stop(self):
        self._stop_event.set()

    def is_event_set(self):
        return self._stop_event.is_set()

    def clear_event(self):
        self._stop_event.clear()

    def start(self):
        raise NotImplementedError("%s::start not implemented" % type(self).name)

    def close(self):
        self.stop()
        if self.socket:
            self.socket.close()
            self.socket = None

    stopped = is_event_set



#!/usr/bin/env python
# Copyright (c) 2019 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.

from lydian.apps.base import BaseApp, exposify
from lydian.utils.network_utils import NamespaceManager


@exposify
class NamespaceApp(object):
    def __init__(self):
        self._ns_manager = NamespaceManager()

    def list_namespaces(self):
        return self._ns_manager.get_all_namespaces()

    def get_namespace(self, namespace):
        return self._ns_manager.get_namespace(namespace)

    def list_namespaces_ips(self):
        return self._ns_manager.get_all_namespaces_ips()

    def discover_namespaces(self):
        self._ns_manager.discover_namespaces()



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.

'''
Traffic Controller App handles all the pre processing and post processing
work for Traffic Generation.
'''
import pickle
import logging

from lydian.apps.base import BaseApp, exposify

import lydian.traffic.core as core
import lydian.traffic.task as task

from lydian.utils.network_utils import NamespaceManager, InterfaceManager, \
        NAMESPACE_INTERFACE_NAME_PREFIXES

from lydian.traffic.manager import ClientManager, ServerManager
from lydian.utils.common import get_mgmt_ifname, get_host_name


log = logging.getLogger(__name__)

@exposify
class TrafficControllerApp(BaseApp):

    def __init__(self, record_queue, rulesApp):
        super(TrafficControllerApp, self).__init__()

        self._recore_queue = record_queue
        self.rules = rulesApp

        # TODO : Following should be consumed through Global Apps actually.
        # Namespace Manager handles all namespace information fetching.
        self._ns_mgr = NamespaceManager()
        # Interface Manager
        self._if_mgr = InterfaceManager()
        # Endpoint - Target Map. A target can be host, Namespace or
        # a container.

        # Host is where this app is running. Based on host,
        # it is decided for a rule if we need to run a client
        # or server.
        ifname = get_mgmt_ifname()
        self._host = self._if_mgr.get_interface(ifname)['address']
        self._update_endpoints_map()

        self._client_mgr = ClientManager(self._recore_queue)
        self._server_mgr = ServerManager()

    @property
    def host(self):
        return self._host

    def _update_endpoints_map(self):
        self._ep_map = {}

        # Update Interfaces on this host
        host_target = core.VMHost(name=get_host_name(),
                                  ip=self.host)

        # To support local traffic.
        self._ep_map['127.0.0.1'] = host_target
        self._ep_map['::1'] = host_target

        for ifname in self._if_mgr.get_all_interfaces():
            if not any([ifname.startswith(x) for x in NAMESPACE_INTERFACE_NAME_PREFIXES]):
                continue
            interface = self._if_mgr.get_interface(ifname)
            self._ep_map[interface['address']] = host_target

        # Update Namespaces on this host.
        for ns_name, ns_interfaces in self._ns_mgr.get_namespace_interface_map().items():
            ns_target = core.NSHost(name=ns_name, ip=self.host)
            for interface in ns_interfaces:
                self._ep_map[interface['address']] = ns_target

    def discover_interfaces(self):
        """ Re/Discovers insterfaces """
        self._if_mgr._discover_interfaces()
        self._ns_mgr.discover_namespaces()
        self._update_endpoints_map()

    def _add_rule_info(self, trule):
        trule.src_target = self._ep_map.get(trule.src)
        trule.dst_target = self._ep_map.get(trule.dst)
        trule.src_host = self.host if trule.src_target else None
        trule.dst_host = self.host if trule.dst_target else None

        if not trule.src_host and not trule.dst_host:
            log.error("Invalid request to add rule (%s) on host %s",
                      trule, self.host)
            return

        # Set Active/Inactive State
        trule.state = getattr(trule, 'state', core.TrafficRule.ACTIVE)

        # Add server on this host if needed. Start server before the client.
        if trule.dst_host:
            self._server_mgr.add_task(trule)

        # Add client on this host if needed.
        if trule.src_host:
            self._client_mgr.add_task(trule)

    def _get_traffic_rule(self, rule):
        """ Rule config. """
        log.info("Processing rule : %s", rule)
        log.info("%s", type(rule).__name__)
        rule = dict(rule)
        log.info("After eval %s", type(rule).__name__)
        trule = core.TrafficRule()

        for key, val in rule.items():
            setattr(trule, key, val)

        self._add_rule_info(trule)
        return trule

    def register_traffic(self, traffic_rules=None):
        try:
            traffic_rules = pickle.loads(traffic_rules)
        except Exception:
            # unpickled data. Let the error be raised later
            # if we cann't process it.
            pass    # unpickled data.
        log.info("Registering Traffic : %r", traffic_rules)
        for rule in traffic_rules:
            # create a rule and add it to database.
            log.info("%s", type(rule).__name__)
            trule = self._get_traffic_rule(rule)
            # self.rules.add(trule)     # TODO : Fix rules app and enable it.
        log.info("Registered Traffic Successfully: %r", traffic_rules)

    def register_rule(self, trule):
        try:
            trule = pickle.loads(trule)
        except Exception:
            # unpickled data. Let the error be raised later
            # if we cann't process it.
            pass    # unpickled data.

        log.info("Registering Traffic : %r", trule)
        self._add_rule_info(trule)
        # self.rules.add(trule)     # TODO : Fix rules app and enableit.
        log.info("Registered Traffic Successfully: %r", trule)
        return trule

    def start(self, ruleid):
        """ Start a Traffic task (again). """
        self._client_mgr.start(ruleid)

    def stop(self, ruleid):
        """ Stop a Traffic task. """
        self._client_mgr.stop(ruleid)

    def close(self):
        self._client_mgr.close()
        self._server_mgr.close()



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.
'''
A simple test case for sending traffic.
'''

import logging
import os
import queue
import time
import unittest
import uuid

from lydian.apps.rules import RulesApp
from lydian.apps.controller import TrafficControllerApp
from lydian.apps.results import Results
from lydian.apps.recorder import RecordManager
from lydian.apps.monitor import ResourceMonitor
from lydian.recorder.wf_client import WavefrontTrafficRecorder, WavefrontResourceRecorder

log = logging.getLogger(__name__)


class TrafficAppTest(unittest.TestCase):
    DB_FILE = 'test_traffic_rules.db'
    MAX_QUEUE_SIZE = 20000

    DUMMY_RULE = {
        'reqid': '%s' % uuid.uuid4(),
        'ruleid': '%s' % uuid.uuid4(),
        'src': '127.0.0.1',
        'dst': '127.0.0.1',
        'protocol': 'TCP',
        'port': 9465,
        'connected': True
    }

    def setUp(self):
        if os.path.exists(self.DB_FILE):
            os.remove(self.DB_FILE)

        # Traffic Records.
        self.traffic_records = queue.Queue(self.MAX_QUEUE_SIZE)
        # Resource records.
        self.resource_records = queue.Queue(self.MAX_QUEUE_SIZE)

        self.rulesApp = RulesApp(db_file=self.DB_FILE)
        self.controller = TrafficControllerApp(self.traffic_records, self.rulesApp)
        self.monitor = ResourceMonitor(self.resource_records)
        self.results = Results()

        self.db_pool = RecordManager(self.traffic_records, self.resource_records)

        self.db_pool.start()
        self.monitor.start()

    def test_traffic(self):
        traffic_rules = [self.DUMMY_RULE]

        self.controller.register_traffic(traffic_rules)

        time.sleep(2)  # Wait for taffic to run for 10 seconds.

        records = self.results.traffic(self.DUMMY_RULE['reqid'])
        assert records, "Traffic results missing"

    def tearDown(self):
        self.controller.close()
        self.db_pool.close()
        self.rulesApp.close()
        os.remove(self.DB_FILE)



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.

import logging
import time

from lydian.apps.base import BaseApp, exposify
from lydian.utils.prep import prep_node
from lydian.controller.client import LydianClient

import lydian.common.consts as consts


log = logging.getLogger(__name__)

_podium = None


def _get_host_ip(host, func_ip=None):
    func = lambda vm: vm.ip
    func_ip = func if func_ip is None else func_ip
    return func_ip(host)


@exposify
class Podium(BaseApp):

    TENNAT_VM_USERNAME = 'root'
    TENNAT_VM_PASSWORD = '!cisco'

    HOST_WAIT_TIME = 4

    def __init__(self, username=None, password=None):
        """
        Podium app for running the show.
        # TODO : Add persistence to Podium app.
        """
        self._primary = True
        self._ep_hosts = {}
        self._ep_username = username or self.TENNAT_VM_USERNAME
        self._ep_password = password or self.TENNAT_VM_PASSWORD

    @property
    def endpoints(self):
        return self._ep_hosts.keys()

    def wait_on_host(self, hostip, wait_time=None):
        wait_time = wait_time or self.HOST_WAIT_TIME
        et = int(time.time() + wait_time)

        while time.time() < et:
            if self.is_host_up(hostip):
                return True

        return False

    def add_endpoints(self, hostip, username=None, password=None):
        """
        Add endpoints from the host, reachable by hostip.
        """
        if hostip in self.endpoints:
            log.info("%s is already added.", hostip)
            return

        username = username or self._ep_username
        password = password or self._ep_password

        try:
            with LydianClient(hostip) as client:
                # fetch regular interfaces
                for iface, ips in client.interface.get_interface_ips_map().items():
                    if not any([iface.startswith(x) for x in
                                consts.NAMESPACE_INTERFACE_NAME_PREFIXES]):
                        continue
                    for ip in ips:
                        self._ep_hosts[ip] = hostip

                # Fetch Namespace Interfaces
                for ip in client.namespace.list_namespaces_ips():
                    self._ep_hosts[ip] = hostip

            self._ep_hosts[hostip] = hostip

        except Exception as err:
            log.error("Error in adding endpoint %s - %r", hostip, err)

    def add_hosts(self, hostip, username=None, password=None):
        username = username or self._ep_username
        password = password or self._ep_password
        try:
            prep_node(hostip, username, password)
            if not self.wait_on_host(hostip):
                log.error("Could not start service on %s", hostip)
            self.add_endpoints(hostip, username, password)
        except Exception as err:
            log.error("Error in preparing host %s - %r", hostip, err)

    def is_host_up(self, hostip):
        try:
            with LydianClient(hostip) as client:
                client.monitor.is_running()
            return True
        except Exception:
            return False

    def get_ep_host(self, epip):
        return self._ep_hosts.get(epip, None)

    def register_traffic(self, intent):
        # TODO : Optimization opportunities
        # Club all requests to same host in one call.
        for rule in intent:
            srchost = self.get_ep_host(rule['src'])
            dsthost = self.get_ep_host(rule['dst'])

            if not srchost:
                log.error("No host found for running traffic from IP : %s",
                          rule['src'])
                continue
            elif not dsthost:
                log.error("No host found for running traffic from IP : %s",
                          rule['dst'])
                continue

            # Start Server before the client.
            with LydianClient(dsthost) as dclient:
                dclient.controller.register_traffic([rule])

            with LydianClient(srchost) as sclient:
                sclient.controller.register_traffic([rule])


def get_podium():
    global _podium
    if not _podium:
        _podium = Podium()

    return _podium


def run_iperf3(src, dst, duration=10, udp=False, bandwidth=None,
               client_args='', server_args='', func_ip=None):
    """
    Run iperf between <src> and <dst> over TCP/UDP for <duration> seconds

    Returns iperf client output

    NOTE: By default the transfer rate is unlimited which consumes high CPU

    Parameters
    ----------
    src: host
        iperf client
    dst: host
        iperf server
    duration: int
        How long iperf should run
    udp: bool
        Whether to run in UDP mode or TCP (default: TCP)
    bandwidth: int
        Limit traffic to this many Mbits/second
    client_args: str
        Additional cli options supported by iperf client
    server_args: str
        Additional cli options supported by iperf server
    """
    _podium = get_podium()
    src_host = _podium.get_ep_host(src)
    dst_host = _podium.get_ep_host(dst)
    with LydianClient(dst_host) as server:
        with LydianClient(src_host) as client:
            try:
                port = server.iperf.start_iperf_server(args=server_args)
                log.info('iperf server: %s is running on port %s', dst_host, port)
                job_id = client.iperf.start_iperf_client(dst_host, port, duration, udp, bandwidth,
                                                         args=client_args)
                job_info = client.iperf.get_client_job_info(job_id)
                log.info('cmd: %s on iperf client running with job id: %d', job_info['cmd'],
                         job_id)
                time.sleep(duration)
                while job_info['state'] == 'running':
                    time.sleep(1)
                    job_info = client.iperf.get_client_job_info(job_id)
                    log.info('iperf client job: %d info %s', job_id, job_info)
                return job_info['result']
            finally:
                if port:
                    server.iperf.stop_iperf_server(port)
                    if job_id:
                        client.iperf.stop_iperf_client(job_id)


def start_pcap(host, pcap_file_name, interface, pcap_args='', func_ip=None):
    """
    Starts packet capture on a requested host.
    """
    with LydianClient(_get_host_ip(host, func_ip)) as client:
        client.pcap.start_pcap(pcap_file_name, interface, pcap_args)


def stop_pcap(host, pcap_file_name, func_ip=None):
    """
    Starts packet capture on a requested host.
    """
    with LydianClient(_get_host_ip(host, func_ip)) as client:
        client.pcap.stop_pcap(pcap_file_name)


def start_resource_monitoring(host, func_ip=None):
    """
    Starts packet capture on a requested host.
    """
    with LydianClient(_get_host_ip(host, func_ip)) as client:
        client.monitor.start()


def stop_resource_monitoring(host, func_ip=None):
    """
    Starts packet capture on a requested host.
    """
    with LydianClient(_get_host_ip(host, func_ip)) as client:
        client.monitor.stop()



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.

import json
import uuid
import time


class Record(object):

    def __init__(self):
        """
        Represents a Base class for recording time series data.
        """
        self._id = None
        self._timestamp = int(time.time())

    def __repr__(self):
        return '%s' % self.toJSON()

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)
    @property
    def timestamp(self):
        return self._timestamp

    def as_dict(self):
        return self.__dict__


class ResourceRecord(Record):

    def __init__(self, syscpu=None, sysmem=None, sysconns=None,
                 lydiancpu=None, lydianmem=None, lydianconns=None):
        super(ResourceRecord, self).__init__()
        self.syscpu = syscpu
        self.sysmem = sysmem
        self.sysconns = sysconns
        self.lydiancpu = lydiancpu
        self.lydianmem = lydianmem
        self.lydianconns = lydianconns


class TrafficRecord(Record):

    def __init__(self):
        super(TrafficRecord, self).__init__()
        self.source = None          # ping source IP Address
        self.destination = None     # ping destination IP Address
        self.protocol = None           # ping traffic type (TCP/UDP/HTTP)
        self.expected = None        # ping expectation (Pass/Fail : True/False)
        self.result = None          # Actual result



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.

import json
import logging
import os
import platform
import socket
import sys

from sql30 import db

from lydian.apps.base import BaseApp


log = logging.getLogger(__name__)
configs = None

#### CONSTANTS ####
SYSTEM = platform.system()
LINUX_OS = True if SYSTEM == 'Linux' else False
MAC_OS = True if SYSTEM == 'Darwin' else False
WIN_OS = True if SYSTEM == 'Windows' else False

# Logging Constants
LINUX_LOG_DIR = os.environ.get('LINUX_LOG_DIR', "/var/log/lydian")
WIN_LOG_DIR = os.environ.get('WIN_LOG_DIR', "C:\\lydian\\log")
if LINUX_OS or MAC_OS:
    LOG_DIR = LINUX_LOG_DIR
elif WIN_OS:
    LOG_DIR = WIN_LOG_DIR
LOG_FILE = "lydian.log"

# Axon Service Constants
AXON_PORT = 5678
LYDIAN_PORT = 5649

# Recorder Constants
WAVEFRONT = 'wavefront'
SQL = 'sql'
ELASTIC_SEARCH = 'elasticsearch'
ELASTIC_SEARCH_PORT = 9200

# # # # # All Configurable Variables set below # # # # #


# Traffic Server Configs
REQUEST_QUEUE_SIZE = 100
PACKET_SIZE = 1024
ALLOW_REUSE_ADDRESS = True


# Env Configs
TEST_ID = os.environ.get('TEST_ID', '')
TESTBED_NAME = os.environ.get('TESTBED_NAME', '')
AXON_PORT = int(os.environ.get('AXON_PORT', 5678))


# Wavefront recorder configs
WAVEFRONT_PROXY_ADDRESS = os.environ.get('WAVEFRONT_PROXY_ADDRESS', None)
WAVEFRONT_SERVER_ADDRESS = os.environ.get('WAVEFRONT_SERVER_ADDRESS', '')
WAVEFRONT_SERVER_API_TOKEN = os.environ.get('WAVEFRONT_SERVER_API_TOKEN', '')
WAVEFRONT_SOURCE_TAG = os.environ.get('WAVEFRONT_SOURCE', socket.gethostname())
WAVEFRONT_REPORT_PERC = float(os.environ.get('WAVEFRONT_REPORT_PERC', 1.0))


# Namespace Configs
NAMESPACE_MODE = os.environ.get("NAMESPACE_MODE", False)
NAMESPACE_MODE = True if NAMESPACE_MODE in ['True', True] else False
NAMESPACE_INTERFACE_NAME_PREFIXES = ["veth", "eth"]


# Recorder Configs
RECORDER = os.environ.get('RECORDER', None)
RECORD_COUNT_UPDATER_SLEEP_INTERVAL = 30
RECORD_UPDATER_THREAD_POOL_SIZE = 50

ELASTIC_SEARCH_SERVER_ADDRESS = os.environ.get('ELASTIC_SEARCH_SERVER_ADDRESS', None)
ELASTIC_SEARCH_SERVER_PORT = os.environ.get(
    'ELASTIC_SEARCH_SERVER_PORT', ELASTIC_SEARCH_PORT)

# # # # # End of Configurable Variables  # # # # #


class ConfigDB(db.Model):
    DB_NAME = 'params.db'
    TABLE = 'config'

    DB_SCHEMA = {
        'db_name': DB_NAME,
        'tables': [
            {
                'name': TABLE,
                'fields': {
                    'param': 'text',
                    'value': 'text',
                    'typename': 'text',
                    },
                'primary_key': 'param'  # avoid duplicate entries.
            }]
        }
    VALIDATE_BEFORE_WRITE = True

class Config(ConfigDB, BaseApp):
    NAME = "CONFIG"

    def __init__(self, db_file=None):
        # Set database name.
        db_name = db_file or self.DB_NAME
        super(Config, self).__init__(db_name=db_name)
        self.table = self.TABLE

        # set params
        self._params = {}

        # Read configs, load from db file.

        self._read_config()
        self.load_from_db()
        self.save_to_db()

    def _read_config(self):
        """
        Reads config from default config file.

        We do not write these configs into the database yet as database
        configs are supposed to overwrite.
        """

        module_name = sys.modules[__name__]
        VARS = [var for var in dir(module_name) if var.isupper()]
        for var in VARS:
            param, val = var, getattr(module_name, var)
            self._params[param] = val

    def _type_handler(self, val, type_name):

        types_map = {
                'int': lambda x: int(x),
                'float': lambda x: float(x),
                'tuple': lambda x: tuple(x),
                'set': lambda x: set(x),
                'bool': lambda x: True if x == 'True' else False,
                'NoneType': lambda x: None,
            }
        val = json.loads(val)

        return types_map[type_name](val) if type_name in types_map else val

    def load_from_db(self):
        """
        Load config params from database file to local cache.
        """
        configs = []
        with ConfigDB() as db:
            configs = db.read(tbl=self.TABLE)
        for key, val, type_name in configs:
            self._params[key] = self._type_handler(val, type_name)

    def save_to_db(self):
        """
        Save config params in local cache to database file.
        """
        with ConfigDB() as db:
            db.table = self.TABLE
            for param, val in self._params.items():
                self._persist_param(param, val, db)

    def get_param(self, param):
        """
        Return the value of a config param. Param is always
        returned from local cache as it is simply a reflector of database file.
        """
        return self._params.get(param, None)

    def set_param(self, param, val, write_to_db=True):
        self._params[param] = val

        if write_to_db:
            with ConfigDB() as db:
                db.table = self.TABLE
                self._persist_param(param, val, db)

    def _persist_param(self, param, val, db_handle):
        """
        Sets a param, val in database file.
        """
        type_name = type(val).__name__
        if isinstance(val, set):
              # json.dumps can't serialize sets. We still return the
              # value as set as type is stored as "set"
            val = list(val)
        record = db_handle.read(param=param)
        if record:
            db_handle.update(condition={'param': param},
                             value=json.dumps(val),
                             typename=type_name)
        else:
            db_handle.write(param=param,
                            value=json.dumps(val),
                            typename=type_name)

    exposed_get_param = get_param
    exposed_set_param = set_param


def get_configs():
    global configs
    if not configs:
        configs = Config()

    return configs


def get_param(param):
    return get_configs().get_param(param)


def set_param(param, val):
    return get_configs().set_param(param, val)



"""
Utilties to temporarily enter linux namespaces.

Demo:

$ ip netns add testns
$ ip link add veth0 type veth peer name veth1
$ ip link set veth1 netns testns
$ ip link list | grep veth
315: veth0: <BROADCAST,MULTICAST> mtu 1500 qdisc noop state DOWN mode DEFAULT group default qlen 1000
$ ip netns exec testns ip link list | grep veth
314: veth1: <BROADCAST,MULTICAST> mtu 1500 qdisc noop state DOWN mode DEFAULT group default qlen 1000
$ python
...
>>> from lydian.utils.nsenter import *
>>> print(os.popen('ip link list | grep veth').read().strip())
315: veth0: <BROADCAST,MULTICAST> mtu 1500 qdisc noop state DOWN mode DEFAULT group default qlen 1000
>>> with namespace('/var/run/netns/testns', 'net'):
...     print(os.popen('ip link list | grep veth').read().strip())
...
314: veth1: <BROADCAST,MULTICAST> mtu 1500 qdisc noop state DOWN mode DEFAULT group default qlen 1000
>>>
>>> print(os.popen('ip link list | grep veth').read().strip())
315: veth0: <BROADCAST,MULTICAST> mtu 1500 qdisc noop state DOWN mode DEFAULT group default qlen 1000
>>>
"""

import contextlib
import ctypes.util
import logging
import os

logger = logging.getLogger(__name__)
LIBC = ctypes.CDLL(ctypes.util.find_library('c'), use_errno=True)
VALID_NAMESPACES = frozenset(['ipc', 'mnt', 'net', 'pid', 'user', 'uts'])


def setns(fd):
    """See http://man7.org/linux/man-pages/man2/setns.2.html"""
    if LIBC.setns(fd, 0) == -1:
        err = ctypes.get_errno()
        raise OSError(err, errno.errorcode[err])


@contextlib.contextmanager
def fdopen(path):
    """Open a read-only fd to <path>"""
    fd = os.open(path, os.O_RDONLY)
    try:
        yield fd
    finally:
        os.close(fd)


@contextlib.contextmanager
def namespace(nspath, nstype):
    """Enter the provided namespace"""
    if nstype not in VALID_NAMESPACES:
        raise ValueError("invalid namespace type %r" % nstype)

    with fdopen("/proc/self/ns/%s" % nstype) as original_ns:
        with fdopen(nspath) as new_ns:
            logger.info("entering %s namespace %s", nstype, nspath)
            setns(new_ns)
        try:
            yield
        finally:
            logger.info("leaving %s namespace %s", nstype, nspath)
            setns(original_ns)



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.

from wavefront_sdk import WavefrontDirectClient

import lydian.apps.config as conf

def _get_wf_sender():
    # Create a sender with:
    # your Wavefront URL
    # a Wavefront API token that was created with direct ingestion permission
    # max queue size (in data points). Default: 50,000
    # batch size (in data points). Default: 10,000
    # flush interval  (in seconds). Default: 1 second
    return WavefrontDirectClient(
        server='https://vmware.wavefront.com',
        token='d9302bfa-89b4-4d7f-995e-830887b9e903')


class WavefrontRecorder(object):

    def __init__(self):
        self._client = _get_wf_sender()
        self._testbed = conf.TESTBED_NAME or ''
        self._testid = conf.TEST_ID or ''


class WavefrontTrafficRecorder(WavefrontRecorder):

    def write(self, record):
        # assert isinstance(trec, TrafficRecord)
        prefix = 'lydian.traffic.' + record.protocol + ".result"
        tags = {
            "datacenter": self._testbed,
            "test_id": self._testbed,
            "reqid": record.reqid,
            "ruleid": record.ruleid,
            "source": record.source,
            "destination": record.destination
            }
        value=1 if record.result else 0
        self._client.send_metric(
                    name=prefix, value=value,
                    timestamp=record.timestamp,
                    source=conf.WAVEFRONT_SOURCE_TAG,
                    tags=tags)


class WavefrontResourceRecorder(WavefrontRecorder):

    def write(self, record):
        # assert isinstance(trec, ResourceRecord)
        prefix = 'lydian.resources.'
        tags = {"datacenter": self._testbed,
                "test_id": self._testid}
        for key, val in record.as_dict().items():
            if key in ['_id', '_timestamp']:
                continue
            metric = prefix + key
            self._client.send_metric(
                    name=metric, value=val,
                    timestamp=record.timestamp,
                    source=conf.WAVEFRONT_SOURCE_TAG, tags=tags)



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.
"""
App for running iperf server or client.
"""
import logging
import random
import time

from lydian.apps.base import BaseApp, exposify
import lydian.apps.console as console

log = logging.getLogger(__name__)

@exposify
class Iperf(console.Console):

    IPERF_BIN = 'iperf3'
    _START_RAND_PORT = 5600
    _END_RAND_PORT = 6000
    _STARTING_JOB_ID = 1000

    def __init__(self):
        super(Iperf, self).__init__()
        # {<port>: <popen obj>}
        self._server_handles = {}

        # {<job_id>: {'popen_obj': <popen obj>,
        #             'state': 'running' | 'done',
        #             'result': < None | JSON output>,
        #             'cmd': <command>
        #             }
        self._client_handles = {}
        self._job_id = None
        self._gen_rand_port = self._gen_rand_port(self._START_RAND_PORT,
                                                  self._END_RAND_PORT)

    def start_iperf_server(self, port=None, args=''):
        """
        Start iperf server on given port. If port is not
        provided, random port will be used.

        :param port: (int) TCP port number (default: None)
        :param args: (str) Additional iperf supported args
        :return:
        port: (int) TCP port number being used for iperf server
        """
        if not port:
            port = next(self._gen_rand_port)
        if self.is_running(port):
            log.info("Server is already running on port: %d"
                     "Skip starting server." % port)
            return port
        cmd = self.IPERF_BIN + ' ' + '-s -p %s' % str(port)
        if args:
            cmd += ' ' + args
        p = self._start_subprocess(cmd)
        # Check server is running
        for _ in range(3):
            if self.is_running(port):
                self._server_handles[port] = p
                log.info("Server is running on port:%d " % port)
                break
            time.sleep(1)
        else:
            msg = "unable to start iperf3 server. check port:%d is available." % port
            log.error(msg)
            raise RuntimeError(msg)
        return port

    def _gen_rand_port(self, start_port, end_port):

        for _ in range(start_port, end_port + 1):
            port = random.randint(start_port, end_port)
            if port not in self._server_handles:
                yield port
        else:
            msg = "Running out of random port between starting port: %d - "\
                             "ending port: %d" % (start_port, end_port)
            log.error(msg)
            raise ValueError(msg)

    def get_server_ports(self):
        """
        Get running server ports.
        :return:
        list of currently running server ports
        """
        return [p for p in self._server_handles.keys()]

    def stop_iperf_server(self, port):
        """
        Stop iperf server for given port.
        """
        proc = self._server_handles.get(port)
        if proc:
            self._kill_subprocess(proc)
            self._server_handles.pop(port)

    def stop_iperf_client(self, job_id):
        """
        Stop iperf client for given job_id.
        """
        job = self._client_handles.get(job_id)
        if job:
            self._kill_subprocess(job.get('popen_obj'))
            self._client_handles.pop(job_id)

    def start_iperf_client(self, dst_ip, dst_port, duration=10, udp=False,
                           bandwidth=None, args=''):
        """
        Start iperf client to given dst_ip and port.

        :param dst_ip: (str) IP address of iperf server
        :param dst_port: (int) iperf server port to connect to
        :param dst_port: (int) iperf server port to connect to
        :param duration: (int) test duration
        :param udp: (bool) enable udp
        :param bandwidth: (int)  limit traffic to as much Mbits/second
        :return:
        job_id: (int)
        """

        cmd = self.IPERF_BIN + ' ' + '-c %s -p %s -t %s' % (str(dst_ip),
                                                            str(dst_port),
                                                            str(duration))
        cmd += ' ' + '--json'
        if udp:
            cmd += " --udp"
            # iperf3 udp default is 1 Mbit/s, this sets it to unlimited / 10Gb/s
            if not bandwidth:
                bandwidth = 10240
        if bandwidth:
            cmd += " --bandwidth %dM" % bandwidth
        if args:
            cmd += ' ' + args
        p = self._start_subprocess(cmd)
        self._job_id = self._STARTING_JOB_ID if not self._job_id \
            else self._job_id + 1

        self._client_handles[self._job_id] = {'popen_obj': p,
                                              'state': 'running',
                                              'result': None,
                                              'cmd': cmd}
        log.info("cmd: %s running with job_id: %s" % (cmd, self._job_id))
        return self._job_id

    def get_client_jobs(self):
        return [j for j in self._client_handles.keys()]

    def get_client_job_info(self, job_id):
        """
        Get client job info on given job id.

        :param job_id: (int)
        :return:
        job_details: (dict)
            # {'popen_obj': <popen obj>,
            #  'state': 'running' | 'done',
            #  'result': < None | JSON output>,
            #  'cmd': <command>
            # }
        """
        if job_id not in self._client_handles:
            log.error("Job ID:%d not found" % job_id)
            return None
        p = self._client_handles[job_id].get('popen_obj')
        if not self._is_alive(p):
            if self._client_handles[job_id].get('result') is None:
                result_json = p.communicate()[0].decode('utf-8')
                result_json = result_json.replace('\n', '').replace('\t', '')
                self._client_handles[job_id]['result'] = result_json
            self._client_handles[job_id]['state'] = 'done'

        return self._client_handles[job_id]

    def stop(self):
        """
        Stop iperf app.
        """
        server_ports = self.get_server_ports()
        client_jobs = self.get_client_jobs()

        for port in server_ports:
            log.info("Stopping iperf server on port %s", port)
            proc = self._server_handles.get(port)
            if proc:
                self._kill_subprocess(proc)

        for job_id in client_jobs:
            log.info("Stopping iperf client process for job: %s" % job_id)
            proc = self._client_handles[job_id].get('popen_obj')
            if proc:
                self._kill_subprocess(proc)

    def is_running(self, port):
        """
        Check iperf server is running on given port
        """
        netstat = self._start_subprocess('netstat -lnp')
        grep = self._start_subprocess('grep :%s' % port, stdin=netstat.stdout)
        netstat.stdout.close()  # Allow netstat to receive a SIGPIPE if grep exits.
        output = grep.communicate()[0].decode('utf-8')
        return self.IPERF_BIN in output



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.

'''
Rules app handles all the operations and management of Traffic
rules. Traffic Rules shall be stored at Primary node as well as
worker nodes. Primary nodes would have all the traffic rules running
in the system at a given point of time. Rules at endpoints would only
be related to that endpoint host.
'''

import logging

from sql30 import db


from lydian.apps.base import BaseApp, exposify
from lydian.traffic.core import TrafficRule

log = logging.getLogger(__name__)


@exposify
class RulesApp(db.Model, BaseApp):
    NAME = "RULES"
    DB_NAME = 'rules.db'
    TABLE = 'rules'
    DB_SCHEMA = {
        'db_name': DB_NAME,
        'tables': [
            {
                'name': TABLE,
                'fields': TrafficRule.SCHEMA,
                'primary_key': 'ruleid'  # avoid duplicate entries.
            }]
        }
    VALIDATE_BEFORE_WRITE = True

    # Valid states
    INACTIVE = 'INACTIVE'
    ACTIVE = 'ACTIVE'

    def __init__(self, db_file=None):

        db_name = db_file or self.DB_NAME

        super(RulesApp, self).__init__(db_name=db_name)
        self._rules = {}    # represents local cache.
        self.table = self.TABLE
        self.load_from_db()

    def get(self, ruleid):
        return self._rules.get(ruleid)

    def add(self, trule, save_to_db=True):
        """
        Adds a rule in local cache and database and returns
        corresponding object.
        """
        if save_to_db:
            if trule.ruleid in self._rules:
                # Local Cache is always in sync with Database.
                self.update(condition={'ruleid': trule.ruleid},
                            **trule.as_dict())
            else:
                self.create(**trule.as_dict())
            # Commit the records.
            self.commit()
        self._rules[trule.ruleid] = trule

    def add_rules(self, trules):
        """ Adds multiple rules. """
        for trule in trules:
            self.add(trule, save_to_db=False)
        self.commit()

    def load_from_db(self):
        """ Loads rules from DB to local file."""
        records = self.read(include_header=True)
        fields = list(records[0])
        for index, rule in enumerate(records):
            if not index:
                continue  # first record is a header
            trule = TrafficRule()

            for key, val in zip(fields, list(rule)):
                setattr(trule, key, val)

            ruleid = getattr(trule, 'ruleid', None)
            if not ruleid:
                self.error("Skipped Invalid rule with no ruleid : %s",
                           trule.__dict__)
            self._rules[ruleid] = trule

    def save_to_db(self):
        """
        Save config params in local cache to database file.
        """
        curr_rule_ids = [x[0] for x in self.read()]  # first value is ruleid
        for ruleid, trule in self._rules.items():
            if ruleid in curr_rule_ids:
                self.update(condition={'ruleid': ruleid}, **trule.as_dict())
            else:
                self.create(**trule.as_dict())

        self.commit()

    def disable(self, ruleid):
        """ Disables a rule. """
        if ruleid not in self._rules:
            log.error("Invalid rule to disable : %s", ruleid)
            return

        self._rules[ruleid].state = self.INACTIVE

        where = {'ruleid': ruleid}
        self.update(condition=where, state=self.INACTIVE)
        self.commit()

    def enable(self, ruleid):
        """ Enables a rule """
        if ruleid not in self._rules:
            log.error("Invalid rule to enable : %s", ruleid)
            return

        self._rules[ruleid].state = self.ACTIVE

        where = {'ruleid': ruleid}
        self.update(condition=where, state=self.ACTIVE)
        self.commit()

    def is_enabled(self, ruleid):
        rule = self._rules.get(ruleid)

        if not rule:
            log.error("Invalid rule id : %s", ruleid)

        return rule.state == self.ACTIVE

    def close(self):
        self.save_to_db()
        super(RulesApp, self).commit()



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.

import importlib
import logging
import os
import zipfile

'''
A simple module for zipping files and creating egg files.
'''


class Zipper(object):
    EXPAND_LINKS = True
    SKIP_COMPILED = True

    def __init__(self, output_file, source=None):
        # Output file with path
        self.dst = output_file

        # input source for zipping
        self.src = source

        self.modules = []

        # skip compiled code
        self.skip_compiled = self.SKIP_COMPILED

        self.clear()

    def clear(self):
        # overwrite existing egg if any.
        with zipfile.ZipFile(self.dst, "w") as zf:
            _ = zf

    def add_module(self, module_name):
        # Import modules like 'pika', 'rpyc'
        # Basically takes care of knowing what all to zip
        try:
            if module_name in self.modules:
                # TODO : message already imported.
                return
            mod = importlib.import_module(module_name)
            modpath, _ = os.path.split(mod.__file__)

            # pack the modules at root level. it makes importing
            # modules from egg directly without modifying PYTHONPATH
            self._write(srcdir=modpath, append=True, atroot=module_name)
        except Exception as err:
            # TODO : Fix the error
            _ = err
            pass

    def add_file(self, filename, arcname=None):
        if not os.path.isfile(filename):
            return
        with zipfile.ZipFile(self.dst, "a") as zf:
            arcname = filename if not arcname else arcname
            zf.write(filename=filename, arcname=arcname)

    def add_dir(self, dirname, atroot=None):
        self._write(srcdir=dirname, append=True, atroot=atroot)

    def _write(self, srcdir=None, append=False, atroot=None):
        mode = "w" if not append else "a"
        srcdir = self.src if srcdir is None else srcdir

        with zipfile.ZipFile(self.dst, mode) as zf:
            for dirname, _, files in os.walk(srcdir):
                for filename in files:
                    # Don't pack .pyc files unless asked to.
                    if self.skip_compiled and filename.endswith('.pyc'):
                        continue

                    _file = os.path.join(dirname, filename)
                    _arcfile = _file
                    if atroot:
                        _arcfile = os.path.join(atroot,
                                                os.path.relpath(_file, srcdir))
                    print ("Writing ", _file, " --> ", _arcfile)
                    zf.write(filename=_file, arcname=_arcfile)


if __name__ == '__main__':
    try:
        z = Zipper(output_file='lydian.egg')
        z.add_module('rpyc')
        z.add_module('sql30')
        # z.add_module('psutil')    # TODO : Resolve this.
        z.add_module('wavefront-sdk-python')
        z.add_module('wavefront-api-client')
        z.add_dir(dirname="lydian", atroot="./lydian/")
        logging.info("Generated sample.egg")
    except Exception as err:
        logging.info("Error in zipping directory")



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.

'''
Base class for all Apps on the platform.
'''

import collections

class BaseApp(object):

    NAME = ''

    @property
    def name(self):
        """
        Returns name of the App.
        """
        assert self.NAME, "Invalid Name for app"
        return self.NAME


def exposify(cls):
    for key in dir(cls):
        val = getattr(cls, key)
        if isinstance(val, collections.Callable) and not key.startswith("_"):
            setattr(cls, "exposed_%s" % (key,), val)
    return cls


#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.

import logging
import queue
import threading

from lydian.traffic.core import TrafficRecord
from lydian.traffic.client import TCPClient, UDPClient
from lydian.traffic.server import TCPServer, UDPServer


log = logging.getLogger(__name__)


class TrafficTask(object):

    CLIENT = 'CLIENT'
    SERVER = 'SERVER'

    def __init__(self, trule):
        self._task = None
        self._type = None
        self._trule = trule   # Traffic Rule
        self._task_thread = None
        self._create_task()

    def is_client(self):
        return self._type == self.CLIENT

    def is_server(self):
        return self._type == self.SERVER

    @property
    def target(self):
        return None

    def _create_task(self):
        if self.target.is_vmhost():
            self._create_vmhost_task()
        elif self.target.is_namespace():
            self._create_namspace_task()
        elif self.target.is_container():
            self._create_container_task()
        else:
            log.error("Unknown target for rule : %s",
                      self._trule)

    def _create_vmhost_task(self):
        raise NotImplementedError("_create_vmhost_task")

    def _create_namspace_task(self):
        raise NotImplementedError("_create_namspace_task")

    def _create_container_task(self):
        raise NotImplementedError("_create_container_task")

    def start(self, blocking=False):
        if blocking:
            self._task.start()
        else:
            self._task_thread = threading.Thread(target=self._task.start)
            self._task_thread.start()

    def _join_thread(self):
        if self._task_thread:
            self._task_thread.join()
            self._task_thread = None

    def stop(self):
        self._task.stop()
        self._join_thread()

    def close(self):
        self._task.close()
        self._join_thread()

    def is_running(self):
        return not self._task.stopped()


class TrafficClientTask(TrafficTask):

    def __init__(self, record_queue, trule):
        self._record_queue = record_queue
        self._type = self.CLIENT
        super(TrafficClientTask, self).__init__(trule)

    @property
    def target(self):
        return self._trule.src_target

    @property
    def record_queue(self):
        return self._record_queue

    def _get_client(self):
        server = self._trule.dst
        port = self._trule.port

        if self._trule.is_TCP():
            return TCPClient(server=server, port=port,
                             handler=self.ping_handler)
        elif self._trule.is_UDP():
            return UDPClient(server=server, port=port)
        else:
            msg = "LYDIAN: Unsupported protocol on rule %s" % self._trule
            raise NotImplementedError(msg)

    def _create_vmhost_task(self):
        self._task = self._get_client()

    def ping_handler(self, payload, data):
        try:
            rec = TrafficRecord()
            rec.source = self._trule.src
            rec.destination = self._trule.dst
            rec.protocol = self._trule.protocol
            rec.port = self._trule.port
            rec.expected = self._trule.connected
            rec.result = not rec.expected ^ (data == payload)
            rec.reqid = self._trule.reqid
            rec.ruleid = self._trule.ruleid
            # log.info("Traffic: %r", rec)
            self.record_queue.put(rec, block=False, timeout=2)
        except queue.Full:
            log.error("Cann't put Traffic Record %r into the queue.",
                      rec)


class TrafficServerTask(TrafficTask):

    def __init__(self, trule):
        self._type = self.SERVER
        super(TrafficServerTask, self).__init__(trule)

    @property
    def target(self):
        return self._trule.dst_target

    def _get_server(self):
        port = self._trule.port
        if self._trule.is_TCP():
            return TCPServer(port=port)
        elif self._trule.is_UDP():
            return UDPServer(port=port)
        else:
            msg = "LYDIAN: Unsupported protocol on rule %s" % self._trule
            raise NotImplementedError(msg)

    def _create_vmhost_task(self):
        self._task = self._get_server()



#!/usr/bin/env python
# Copyright (c) 2019 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.
import errno
import pickle
import socket
import time

import rpyc
import lydian.common.consts as consts


rpyc.core.protocol.DEFAULT_CONFIG['allow_pickle'] = True


class Manager(object):
    def __init__(self, client):
        self._client = client


class TrafficManger(Manager):

    def delete_traffic_rules(self, endpoint=None):
        self._client.traffic.delete_traffic_rules(endpoint)

    def add_server(self, protocol, port, endpoint, namespace=None):
        self._client.traffic.add_server(protocol, port, endpoint, namespace)

    def start_clients(self):
        self._client.traffic.start_clients()

    def get_traffic_rules(self, endpoint=None):
        return self._client.traffic.get_traffic_rules(endpoint)

    def register_traffic(self, traffic_rules):
        return self._client.traffic.register_traffic(traffic_rules)

    def unregister_traffic(self, traffic_rules):
        self._client.traffic.unregister_traffic(traffic_rules)

    def list_servers(self):
        return self._client.traffic.list_servers()

    def get_server(self, protocol, port):
        return self._client.traffic.get_server(protocol, port)

    def stop_servers(self, namespace=None):
        self._client.traffic.stop_servers(namespace)

    def start_servers(self, namespace=None):
        return self._client.traffic.start_servers(namespace)

    def stop_server(self, protocol, port, namespace=None, endpoint=None):
        self._client.traffic.stop_server(protocol, port, namespace, endpoint)

    def stop_client(self, namespace=None, endpoint=None):
        self._client.traffic.stop_client(namespace, endpoint)

    def stop_clients(self, namespace=None):
        self._client.traffic.stop_clients(namespace)

    def rediscover_namespaces(self):
        self._client.traffic.rediscover_namespaces()


class StatsManager(Manager):

    def get_traffic_stats(self, start_time=None, end_time=None):
        return self._client.stats.get_traffic_stats(
            start_time=start_time, end_time=end_time)

    def get_avg_latency(self, start_time=None, end_time=None):
        return self._client.stats.get_avg_latency(
            start_time=start_time, end_time=end_time)

    def get_failure_count(self, start_time=None, end_time=None,
                          destination=None, port=None, source=None):
        return self._client.stats.get_failure_count(
            start_time=start_time, end_time=end_time,
            destination=destination, port=port, source=source)

    def get_success_count(self, start_time=None, end_time=None,
                          destination=None, port=None, source=None):
        return self._client.stats.get_success_count(
            start_time=start_time, end_time=end_time,
            destination=destination, port=port, source=source)

    def get_failures(self, start_time=None, end_time=None,
                     destination=None, port=None, source=None):
        return self._client.stats.get_failures(
            start_time=start_time, end_time=end_time,
            destination=destination, port=port, source=source)

    def get_successes(self, start_time=None, end_time=None,
                      destination=None, port=None, source=None):
        return self._client.stats.get_successes(
            start_time=start_time, end_time=end_time,
            destination=destination, port=port, source=source)


class NamespaceManager(Manager):

    def list_namespaces(self):
        return self._client.namespace.list_namespaces()

    def get_namespace(self, name):
        return self._client.namespace.get_namespace(name)

    def list_namespaces_ips(self):
        return self._client.namespace.list_namespaces_ips()

    def discover_namespaces(self):
        self._client.namespace.discover_namespaces()


class InterfaceManager(Manager):

    def list_interfaces(self):
        return self._client.interface.list_interfaces()

    def get_interface(self, name, ip=None):
        return self._client.interface.get_interface(name, ip)

    def get_all_ips(self):
        return self._client.interface.get_all_ips()

    def get_ips_by_interface(self, interface):
        return self._client.interface.get_ips_by_interface(interface)

    def get_interface_ips_map(self):
        return pickle.loads(self._client.interface.get_interface_ips_map())


class ResourceMonitorManager(Manager):

    def start(self):
        self._client.monitor.start()

    def stop(self):
        self._client.monitor.stop()

    def is_running(self):
        return self._client.monitor.is_running()


class TCPDumpManager(Manager):

    def start_pcap(self, dst_file, interface='eth0', args=''):
        self._client.tcpdump.start_pcap(dst_file, interface, args)

    def stop_pcap(self, dst_file):
        self._client.tcpdump.stop_pcap(dst_file)

    def is_running(self, dst_file):
        return self._client.tcpdump.is_running(dst_file)


class IperfManager(Manager):

    def start_iperf_server(self, port=None, args=''):
        return self._client.iperf.start_iperf_server(port, args)

    def stop_iperf_server(self, port):
        self._client.iperf.stop_iperf_server(port)

    def start_iperf_client(self, dst_ip, dst_port, duration=10,
                           udp=False, bandwidth=None, args=''):
        return self._client.iperf.start_iperf_client(dst_ip, dst_port,
                                                     duration, udp,
                                                     bandwidth, args)

    def stop_iperf_client(self, job_id):
        self._client.iperf.stop_iperf_client(job_id)

    def get_server_ports(self):
        return self._client.iperf.get_server_ports()

    def get_client_jobs(self):
        return self._client.iperf.get_client_jobs()

    def get_client_job_info(self, job_id):
        return self._client.iperf.get_client_job_info(job_id)

    def is_running(self, port):
        return self._client.iperf.is_running(port)


class ScapyManager(Manager):

    def run(self, code):
        return self._client.scapy.execute(code)

    def get_console(self, params):
        return self._client.scapy.get_console(params)

    def close_session(self, session_id=None):
        self._client.scapy.close_session()


class ConfigManager(Manager):

    def get_param(self, param):
        return self._client.configs.get_param(param)

    def set_param(self, param, val):
        return self._client.configs.set_param(param, val)


class ResultsManager(Manager):

    def traffic(self, reqid):
        return self._client.results.traffic(reqid)


class TrafficControllerManager(Manager):

    def register_traffic(self, traffic_rules):
        traffic_rules = pickle.dumps(traffic_rules)
        return self._client.controller.register_traffic(traffic_rules)

    def register_rule(self, trule):
        trule = pickle.dumps(trule)
        return self._client.controller.register_rule(trule)


class LydianClient(object):
    """
    Top level object to access Lydian API
    """

    def __init__(self, host, port=consts.LYDIAN_PORT,
                 proxy_host=None, request_timeout=180,
                 retry_count=5, sleep_interval=.5):
        """ Initialization of Client object

        :param host: the host IP where lydian service is running
        :type host: str
        :param port: port on which lydian is listening
        :type port: int
        :param proxy_host: ip of the proxy host
        :type proxy_host: str
        :param request_timeout: rpyc request timeout
        :type request_timeout: int
        :param retry_count: no of retries in case of connection failure
        :type retry_count: int
        :type sleep_interval: number of seconds to sleep between retries
        """
        self._host = host
        self._port = port
        self._proxy_host = proxy_host
        self.rpc_client = None
        self._connect(retry_count, sleep_interval, request_timeout)

        # Interface / Namespaces
        self.namespace = NamespaceManager(self.rpc_client.root)
        self.interface = InterfaceManager(self.rpc_client.root)

        # Traffic Apps
        self.traffic = TrafficManger(self.rpc_client.root)
        self.controller = TrafficControllerManager(self.rpc_client.root)

        # Results / Query Apps
        self.stats = StatsManager(self.rpc_client.root)
        self.results = ResultsManager(self.rpc_client.root)

        # Resource Monitor
        self.monitor = ResourceMonitorManager(self.rpc_client.root)
        # Packet Capture
        self.pcap = TCPDumpManager(self.rpc_client.root)
        # IPerf Traffic
        self.iperf = IperfManager(self.rpc_client.root)

        self.scapy = ScapyManager(self.rpc_client.root)
        # Params / Configs
        self.configs = ConfigManager(self.rpc_client.root)

    def _connect(self, retry_count, sleep_interval, request_timeout):

        # Client will wait for request_timeout seconds in case
        # server takes long time to response, default timeout
        # is 30 seconds which is small for few lydian APIs
        conf = {"sync_request_timeout": request_timeout}
        failure_count = 0
        last_exception = None
        while failure_count < retry_count:
            failure_count += 1
            try:
                if self._proxy_host:
                    conn = rpyc.classic.connect(self._proxy_host)
                    self.rpc_client = conn.modules.rpyc.connect(
                        self._host, self._port, config=conf)
                else:
                    self.rpc_client = rpyc.connect(
                        self._host, self._port, config=conf)
                break
            except socket.timeout as e:
                last_exception = e
                time.sleep(sleep_interval)
            except socket.error as e:
                last_exception = e
                errno = e.errno
                if errno == errno.ECONNRESET or errno == errno.ECONNREFUSED:
                    time.sleep(sleep_interval)
                else:
                    raise
            except Exception:
                raise

        if self.rpc_client is None:
            raise last_exception

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.rpc_client.close()



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.

import logging
import queue
import threading

from lydian.apps.base import BaseApp, exposify
from lydian.recorder.wf_client import WavefrontTrafficRecorder, WavefrontResourceRecorder
from lydian.traffic.core import TrafficRecord
from sql30 import db


log = logging.getLogger(__name__)


class TrafficRecordDB(db.Model):
    DB_NAME = 'traffic.db'
    TABLE = 'traffic'

    DB_SCHEMA = {
        'db_name': DB_NAME,
        'tables': [
            {
                'name': TABLE,
                'fields': {
                    'timestamp': 'text',
                    'reqid': 'text',
                    'ruleid': 'text',
                    'source': 'text',
                    'destination': 'text',
                    'protocol': 'text',
                    'port': 'text',
                    'expected': 'text',
                    'result': 'text'
                    },
            }]
        }
    VALIDATE_BEFORE_WRITE = True



class TrafficRecorder(TrafficRecordDB):
    NAME = "TRAFFIC_RECORDER"

    MAXSIZE = 3000
    FLUSH_FREQ = 3  # every 3 seconds.

    def __init__(self, db_file=None):
        # Set database name.
        db_name = db_file or self.DB_NAME
        super(TrafficRecorder, self).__init__(db_name=db_name)
        self._fields = self._get_fields(self.TABLE)

    def write(self, trec):
        # TODO : Create a pool of records and flush them periodically instead.
        if isinstance(trec, TrafficRecord):
            with TrafficRecordDB() as db:
                values = {k: v for k, v in trec.as_dict().items() if k in self._fields}
                values['timestamp'] = trec.timestamp
                db.write(tbl=self.TABLE, **values)


@exposify
class RecordManager(BaseApp):
    """
    This class act as a deamon to read traffic record queue and to
    write record to the db recorder provided
    """
    RECORD_UPDATER_THREAD_POOL_SIZE = 2
    RESOURCE_RECORD_REPORT_FREQ = 4
    TRAFFIC_RECORD_REPORT_FREQ = 4

    def __init__(self, traffic_records, resource_records):
        self._traffic_recorders = [
            TrafficRecorder(),
            WavefrontTrafficRecorder()
            ]
        self._resource_recorders = [
            WavefrontResourceRecorder()
        ]
        self._traffic_records = traffic_records
        self._resource_records = resource_records
        self._stopped = threading.Event()
        self._handlers = []

    def stopped(self):
        return self._stopped.is_set()

    def stop(self):
        self._stopped.set()

    def _traffic_record_handler(self):
        while not self._stopped.is_set():
            try:
                # TODO : Add logic for flush interval / buffer.
                t_record = self._traffic_records.get(timeout=self.TRAFFIC_RECORD_REPORT_FREQ)
                for recorder in self._traffic_recorders:
                    recorder.write(t_record)
            except queue.Empty:
                pass
            except Exception as err:
                log.error("Error in handling Traffic records : %r", err)

    def _resource_record_handler(self):
        while not self._stopped.is_set():
            try:
                # TODO : Add logic for flush interval / buffer.
                t_record = self._resource_records.get(timeout=self.RESOURCE_RECORD_REPORT_FREQ)
                for recorder in self._resource_recorders:
                    recorder.write(t_record)
            except queue.Empty:
                pass
            except Exception as err:
                log.error("Error in handling Resource records %r", err)

    def start(self, blocking=False):
        self._stopped.clear()

        # Traffic Records Handler
        thandler = threading.Thread(target=self._traffic_record_handler)
        self._handlers.append(thandler)

        # Resource Record handler
        rhandler = threading.Thread(target=self._resource_record_handler)
        self._handlers.append(rhandler)

        for handler in self._handlers:
            handler.start()

    def close(self):
        self.stop()
        _ = [h.join() for h in self._handlers]



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.
# Author : Fred Grim, Vipin Sharma
"""
A standalone utility for connecting to hosts through paramiko over SSH.
"""
import collections
import io
import logging
import os
import pkgutil
import platform
import time

import paramiko


FakeProc = collections.namedtuple('FakeProc', ['stdin', 'stdout', 'stderr'])

def is_py3():
    return platform.python_version().startswith('3')


class Host(object):
    """
    A host that implements the minimum required interface for copying the
    egg file over to a remote machine.

    Parameters
    ----------
    host : string
        The host to connect to
    user : string, optional
        The username to connect as
    passwd : string, optional
        The password for the connecting user
    keyfile : string, optional
        A local keyfile that can authenticate this against this host
    connect : bool, optional
        Connect on instantiation ? or no ?
    """
    AUTOMATION_KEY_NAME = 'automation_rsa'
    REPLY_BUFFER = 2 ** 11  # bytes
    CONNECT_RETRY_LIMIT = 5
    CONNECT_RETRY_SLEEP = 30
    log = logging.getLogger(__name__)

    def __init__(self, host, user=None, passwd=None, keyfile=None,
                 connect=True):
        self.__host = host
        self.__ssh = paramiko.SSHClient()
        self.__ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self._connected = False
        self.__connect_kwargs = {'look_for_keys': False}
        if user is not None:
            self.__connect_kwargs['username'] = user
        if passwd is not None:
            self.__connect_kwargs['password'] = passwd
        if keyfile is not None:
            self.__connect_kwargs['key_filename'] = keyfile
        else:
            kfile = os.path.join('remote_data', self.AUTOMATION_KEY_NAME)
            kfile_abspath = os.path.join(
                os.path.dirname(os.path.abspath(__file__)), kfile)
            if os.path.exists(kfile_abspath):
                keydata = pkgutil.get_data(__name__, kfile)
                pkey_file = io.BytesIO(keydata)
                from_pkey_cb = paramiko.rsakey.RSAKey.from_private_key
                self.__connect_kwargs['pkey'] = from_pkey_cb(pkey_file)
        if connect:
            self.ssh_connect()

    def ssh_connect(self):
        attempts, ssh_exc = 0, None
        while attempts < self.CONNECT_RETRY_LIMIT:
            try:
                self.__ssh.connect(self.__host, **self.__connect_kwargs)
                self._connected = True
                return
            except Exception as exc:
                ssh_exc = exc
                self.log.error('Re-connecting to %s due to %s' % (
                    self.__host, ssh_exc.message))
                time.sleep(self.CONNECT_RETRY_SLEEP)
                attempts += 1
        self.log.error('Connecting to %s failed like %s' % (
            self.__host, ssh_exc.message))

    def get_sftp_connection(self):
        attempts = 0
        sftp_client = None
        while attempts < self.CONNECT_RETRY_LIMIT:
            try:
                sftp_client = self.__ssh.open_sftp()
                break
            except Exception as exc:
                self.log.warn("Retrying after sftp failure %s" % exc)
                attempts += 1
        return sftp_client

    @property
    def connected(self):
        return self._connected

    @property
    def ip(self):
        """
        The original host supported this.
        """
        return self.__host

    def put_file(self, src, dst=None):
        """
        This puts the file on the remote host

        Parameters
        ----------
        src : string
            The local file
        dst : string, optional
            The destination path.  If it none the file is dumped in /tmp
        """
        if dst is None:
            dst = os.path.join('/tmp', os.path.basename(src))
        self.log.debug('%s put_file %s %s' % (self.__host, src, dst))
        sftp_client = self.get_sftp_connection()
        try:
            sftp_client.put(src, dst)
        finally:
            sftp_client.close()

    def get_file(self, src, dst=None):
        """
        This gets the file from the remote host

        Parameters
        ----------
        src : string
            The remote file
        dst : string, optional
            The destination path.  If it none the file is dumped in /tmp
        """
        if dst is None:
            dst = os.path.join('/tmp', os.path.basename(src))
        self.log.debug('%s get_file %s %s' % (self.__host, src, dst))
        sftp_client = self.get_sftp_connection()
        try:
            sftp_client.get(src, dst)
        finally:
            sftp_client.close()

    def check_file_exists(self, path):
        """
        Check that a file exists on the remote node

        Parameters
        ----------
        path : string
            The remote path to check
        """
        self.log.debug('%s check_file_exists %s' % (self.__host, path))
        sftp_client = self.get_sftp_connection()
        try:
            return sftp_client.stat(path).st_mtime > 0
        except IOError:
            return False
        finally:
            sftp_client.close()

    def req_call(self, cmd, msg=None, expected_ret=None):
        """
        This mimics the dynamics of req_call

        Parameters
        ----------
        cmd : string
            The command to call
        msg : string, optional
            ignored
        expected_ret : string, optional
            ignored
        """
        _, _ = msg, expected_ret  # pylint
        self.log.debug('%s req_call %s' % (self.__host, cmd))
        attempts, ssh_exc = 0, None
        chan = None
        while attempts < self.CONNECT_RETRY_LIMIT:
            try:
                chan = self.__ssh.get_transport().open_session()
                break
            except Exception as exc:
                ssh_exc = exc
                time.sleep(self.CONNECT_RETRY_SLEEP)
                attempts += 1
                self.log.error('Re-connecting to %s failed like %s' % (
                    self.__host, ssh_exc.message))

        if chan is not None:
            try:
                chan.exec_command(cmd)
                if chan.recv_exit_status() != 0:
                    raise ValueError(
                        'Nonzero exit status on %s stderr = %s' % (
                            cmd, chan.recv_stderr(self.REPLY_BUFFER)))
                reply = ''
                while True:
                    chunk = chan.recv(self.REPLY_BUFFER)
                    chunk = chunk.decode('utf-8') if is_py3() else chunk
                    if chunk == '':
                        break
                    else:
                        reply += chunk 
                return reply
            finally:
                    chan.close()

    @staticmethod
    def check_return(proc):
        """
        noop
        """
        pass

    def delete_file(self, path, force=False, check=True):
        """
        The removes a file on the host

        Parameters
        ----------
        path : string
            The path to remove
        force : boolean, optional
            Force the removal, defaults to rais eexception on fail
        check : boolean, optional
            Check that the file exists
        """
        _ = check  # make pylint happy
        self.log.debug('%s delete_file %s' % (self.__host, path))
        sftp_client = self.get_sftp_connection()
        try:
            sftp_client.remove(path)
        except IOError:
            if force:
                pass
            else:
                raise
        finally:
            sftp_client.close()

    call = req_call

    def run_command(self, cmd):
        """
        Run a command and return something that looks like the results of a
        subprocess.Popen

        Parameters
        ----------
        cmd : string
            The command to run
        """
        chan = self.__ssh.get_transport().open_session()
        try:
            chan.exec_command(cmd)
            stdout, stderr = io.BytesIO(), io.BytesIO()
            _ = chan.recv_exit_status()
            for callback, fobj in [(chan.recv, stdout),
                                   (chan.recv_stderr, stderr)]:
                while True:
                    chunk = callback(self.REPLY_BUFFER)
                    _chunk = chunk.decode('utf-8') if is_py3() else chunk
                    if _chunk == '':
                        break
                    else:
                        fobj.write(chunk)
                fobj.seek(os.SEEK_SET)
            return FakeProc(io.BytesIO(), stdout, stderr)
        finally:
            chan.close()

    req_command = run_command

    def wait_ready(self):
        """
        Wait until the command is complete. Noop
        """
        return True

    def close(self):
        self.__ssh.close()
        self._connected = False

    def __str__(self):
        return '<%s %s>' % (self.__class__.__name__, self.ip)

    def __repr__(self):
        return '<%s %s at 0x%x>' % (
            self.__class__.__name__, self.ip, id(self))

    def __enter__(self):
        if not self.connected:
            self.ssh_connect()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()


class LocalHost(Host):
    """
    This just runs commands on the localhost
    """

    def __init__(self, user=None, passwd=None, keyfile=None):
        super(LocalHost, self).__init__(
            '127.0.0.1', user=user, passwd=passwd, keyfile=keyfile)




#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.

import logging
import unittest
import uuid

from lydian.traffic.core import TrafficRule, VMHost
from lydian.traffic.manager import ClientManager, ServerManager


log = logging.getLogger(__name__)


class TrafficManagerTest(unittest.TestCase):

    def test_traffic_manager(self):
        """ """
        clientMgr = ClientManager()
        serverMgr = ServerManager()

        rule = TrafficRule()    # Dummy Rule

        rule.ruleid = '%s' % uuid.uuid4()
        rule.src = '127.0.0.1'
        rule.dst = '127.0.0.1'
        rule.port = '3655'
        rule.protocol = 'TCP'

        vm_host = VMHost(name='this_host', ip='127.0.0.1')
        rule.src_host = vm_host
        rule.dst_host = vm_host

        # TEST 1
        # Targets must be specified on Traffic Rule for Traffic Manager
        # to be able to create Client/Servers.
        with self.assertRaises(AttributeError) as err:
            clientMgr.add_task(rule)
            self.assertIn(err.args, "has no attribute 'src_target'")

        # TEST 2:
        # Fix targets and host
        rule.src_host = '127.0.0.1'
        rule.dst_host = '127.0.0.1'
        rule.src_target = vm_host
        rule.dst_target = vm_host

        serverMgr.add_task(rule)
        clientMgr.add_task(rule)

        # TEST 3:
        # Adding another server for same rule shouldn't cause error.
        serverMgr.add_task(rule)
        clientMgr.add_task(rule)
        self.assertEqual(clientMgr.num_tasks(), 1)
        self.assertEqual(serverMgr.num_tasks(), 1)

        # TEST 4:
        # Pretend to be different rule with same configuration.
        rule.ruleid = '%s' % uuid.uuid4()   # pretend to be a different rule.

        # Adding another server for different rule won't cause error and
        # number of active tasks still should be one.
        serverMgr.add_task(rule)
        # Creating a new client for same config but different rule is OK.
        clientMgr.add_task(rule)
        self.assertEqual(clientMgr.num_tasks(), 2)  # 2 clients now.
        self.assertEqual(serverMgr.num_tasks(), 1)  # but only 1 server.



#!/usr/bin/env python
# Copyright (c) 2019 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.

import collections
import logging
# from multiprocessing import Queue
from queue import Queue

import rpyc
from rpyc.utils.server import ThreadPoolServer


from lydian.apps.config import get_configs
from lydian.apps.interface import InterfaceApp
from lydian.apps.iperf import Iperf
from lydian.apps.monitor import ResourceMonitor
from lydian.apps.namespace import NamespaceApp
from lydian.apps.tcpdump import TCPDump
from lydian.apps.controller import TrafficControllerApp
from lydian.apps.recorder import RecordManager
from lydian.apps.results import Results
from lydian.apps.rules import RulesApp

import lydian.utils.logger as logger
import lydian.apps.config as config
import lydian.common.consts as consts

rpyc.core.protocol.DEFAULT_CONFIG['allow_pickle'] = True


class LydianServiceBase(rpyc.Service):

    RPYC_PROTOCOL_CONFIG = rpyc.core.protocol.DEFAULT_CONFIG

    def __init__(self):
        super(LydianServiceBase, self).__init__()

    def on_connect(self, conn):
        print("Connected to %r", conn)

    def on_disconnect(self, conn):
        print("Disconnected from %r", conn)


class LydianService(LydianServiceBase):
    RECORD_QUEUE_SIZE = 50000

    EXPOSED = [
        'monitor',
        'namespace',
        'rules',
        'interface',
        'iperf',
        'tcpdump',
        'results',
        'controller'
    ]

    def __init__(self):
        super(LydianService, self).__init__()

        self._traffic_records = Queue(self.RECORD_QUEUE_SIZE)
        self._resource_records = Queue(self.RECORD_QUEUE_SIZE)

        self.recorder = RecordManager(self._traffic_records,
                                      self._resource_records)

        self.namespace = NamespaceApp()
        self.interface = InterfaceApp()
        self.rules = RulesApp()
        self.controller = TrafficControllerApp(self._traffic_records,
                                               self.rules)
        self.monitor = ResourceMonitor(self._resource_records)
        self.tcpdump = TCPDump()
        self.iperf = Iperf()
        self.results = Results()

        self.expose()

    def expose(self):
        for key in self.EXPOSED:
            setattr(self, 'exposed_'+ key, getattr(self, key))

    def start(self):
        try:
            self.monitor.start()
            self.recorder.start()
        except Exception as err:
            self.logger.exception("Error in starting Services %r", err)

    def stop(self):
        try:
            self.monitor.stop()
            self.recorder.stop()
        except Exception as err:
            self.logger.exception("Error while stopping Services %r", err)


class LydianController(object):

    def __init__(self):
        self.lydian_port = consts.LYDIAN_PORT
        self.service = LydianService()
        self.protocol_config = self.service.RPYC_PROTOCOL_CONFIG
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.WARN)
        self.lydian_service = ThreadPoolServer(
            self.service,
            port=self.lydian_port,
            reuse_addr=True,
            protocol_config=self.protocol_config,
            logger=self.logger, nbThreads=50)

    def start(self):
        self.service.start()
        self.lydian_service.start()

    def stop(self):
        self.service.stop()
        self.lydian_service.close()


def main():
    logger.setup_logging(log_dir=config.LOG_DIR,
                         log_file=config.LOG_FILE)
    lydian_service = LydianController()
    lydian_service.start()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.

import socket
import time

from lydian.traffic.connection import Connection


class Server(Connection):

    def __init__(self, port=None, max_conns=None,
                 handler=None, verbose=False, ipv6=False):
        """
        A basic TCP Server connection listener with default echo reply
        message handler.
        """
        self.host = ''   # blank on server side
        self.port = self.DEFAULT_TCP_SERVER_PORT if port is None else port
        self.max_conns = self.MAX_CONNS if max_conns is None else max_conns
        self._handler = self.echo_handler if handler is None else handler

        super(Server, self).__init__(verbose=verbose)

        # Don't create a socket until we are ready to bind.
        self.socket = None
        self._ipv6 = ipv6

    def echo_handler(self, payload):
        raise NotImplementedError("Handler not implemented in %s" % self.__class__.__name__)


class TCPServer(Server):

    def _create_socket(self):
        """
        Returns a simple TCP server socket.
        """
        sock_type = socket.AF_INET6 if self._ipv6 else socket.AF_INET
        self.socket = socket.socket(sock_type, socket.SOCK_STREAM)

    def start(self):
        """
        API to start server at the requested port and other settings.
        """
        try:
            # Create a socket
            self._create_socket()

            # Bind ot the port
            self.socket.bind((self.host, self.port))

            # accept call from client
            self.socket.listen(self.max_conns)

            if self.verbose:  # TODO : more stringent check
                self.log.info(
                    "TCP Server started on %s:%s",
                    self.host,
                    self.port)

            while not self.is_event_set():
                conn, addr = self.socket.accept()
                if self.verbose:
                    msg = "Connection request received from: %s:%s" % (addr[0], addr[1])
                    self.log.info(msg)
                self._handler(conn)
        except Exception as err:
            _ = err
            pass
        finally:
            self.close()

    def echo_handler(self, conn):
        """
        A simple default echo message handler.
        """
        data = conn.recv(self.MAX_PAYLOAD_SIZE)
        conn.send(data)  # send same data back as echo
        conn.close()    # close immediately.


class UDPServer(Server):

    def _create_socket(self):
        """
        Returns a simple TCP server socket.
        """
        sock_type = socket.AF_INET6 if self._ipv6 else socket.AF_INET
        self.socket = socket.socket(sock_type, socket.SOCK_DGRAM)

    def start(self):
        """
        API to start server at the requested port and other settings.
        """
        # Create a socket
        self._create_socket()

        # Bind ot the port
        self.socket.bind((self.host, self.port))

        if self.verbose:  # TODO : more stringent check
            self.log.info("UDP Server started on %s:%s", self.host, self.port)

        while not self.is_event_set():
            data, addr = self.socket.recvfrom(self.MAX_PAYLOAD_SIZE)
            if self.verbose:
                msg = "Connection request received from: %s:%s" % (addr[0], addr[1])
                self.log.info(msg)
            self._handler(data, addr)

        self.socket.close()

    def echo_handler(self, data, addr):
        """
        A simple default echo message handler.
        """
        if data:
            self.socket.sendto(data, addr)  # send same data back as echo


#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.

import logging
import os
import tempfile


import lydian.apps.console as console

log = logging.getLogger(__name__)

def generate_egg(dest_dir, egg_name='lydian.egg'):
    """
    Packs lydian egg and places it in dest_dir.
    """

    reqs = [
            'rpyc==4.0.2',
            'sql30>=0.1.4',
            'psutil==5.6.6',
            'scapy==2.4.3',
            'wavefront-sdk-python>=1.1.1',
            'wavefront-api-client==2.33.15'
            ]
    commands = [
        'virtualenv -p python3 .',
        'pip install -r requirements.txt',
        'python -mzipper'
        ]

    with tempfile.mkstemp(prefix='lydian', dir='/tmp') as tdir:

        # Create requirements file.
        with open(os.path.join(tdir, 'requirements.txt'), 'w+') as fp:
            for req in req:
                fp.write(req + '\n')

        # Create Egg packing script.
        with open(os.path.join(tdir, 'run.sh'), 'w+') as fp:
            for line in commands:
                fp.write(line + '\n')

        zipper_path = os.path.join(
                os.path.basepath(os.path.abspath(__file__)),
                'zipper.py')

        dest_dir = os.path.join(os.getcwd(), dest_dir, egg_name)

        with console.Console(dir=tdir) as prompt:
            prompt.run('cp %s .' % zipper_path)
            prompt.run('sh -x run.sh')
            try:
                assert os.path.exists(os.path.join(tdir, egg_name))
                log.info("Created lydian egg")
                prompt.run('cp lydian.egg %s' % dest_dir)
            except AssertionError:
                log.error("Error in creating lydian egg")



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.
'''
Generic Results app. Gives results for traffic operations.
'''
import logging

from lydian.apps.base import BaseApp, exposify
from lydian.apps.recorder import TrafficRecordDB


log = logging.getLogger(__name__)


@exposify
class Results(BaseApp):

    def traffic(self, reqid):
        with TrafficRecordDB() as db:
            return db.read(tbl=db.TABLE, reqid=reqid)



import logging
import os

import lydian.common.consts as consts


def create_log_dir(log_dir):
    """
    Create Log directory
    """
    try:
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
    except Exception as e:
        # TODO Change the exception type
        raise RuntimeError(
            "Failed to create log directory %s due to %s" % (log_dir, e))


def setup_logging(log_dir=None, log_file=None):
    """
    Sets up Logging handlers and other environment.
    """
    log_dir = log_dir if log_dir else consts.LOG_DIR
    log_file = log_file if log_file else consts.LOG_FILE
    create_log_dir(log_dir)

    log_file_name = os.path.join(log_dir, log_file)
    log_formatter = logging.Formatter(
        '%(asctime)s::%(levelname)s::%(threadName)s::'
        '%(module)s[%(lineno)04s]::%(message)s')
    root_logger = logging.getLogger()
    # if root_logger.handlers:
    #    return
    root_logger.setLevel(logging.INFO)
    file_handler = logging.FileHandler(log_file_name)
    file_handler.setFormatter(log_formatter)
    root_logger.addHandler(file_handler)
    return root_logger



#!/usr/bin/env python
# Copyright (c) 2019 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.

import pickle

from lydian.apps.base import BaseApp, exposify
from lydian.utils.network_utils import InterfaceManager


@exposify
class InterfaceApp(BaseApp):
    def __init__(self):
        self._if_mngr = InterfaceManager()

    def list_interfaces(self):
        return self._if_mngr.get_all_interfaces()

    def get_interface(self, name, ip=None):
        return self._if_mngr.get_interface(name, ip)

    def get_all_ips(self):
        return self._if_mngr.get_all_ips()

    def get_ips_by_interface(self, interface):
        return self._if_mngr.get_ips_by_interface(interface)

    def get_interface_ips_map(self):
        result = {}
        for iface in self.list_interfaces():
            result[iface] = self.get_ips_by_interface(iface)
        return pickle.dumps(result)



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.
'''
A simple test case for sending traffic.
'''

import logging
import os
import queue
import time
import unittest
import uuid

from lydian.controller.rpyc_controller import LydianController
from lydian.controller.client import LydianClient

log = logging.getLogger(__name__)


class TrafficAppTest(unittest.TestCase):
    DB_FILE = 'test_traffic_rules.db'
    MAX_QUEUE_SIZE = 20000

    DUMMY_RULE = {
        'reqid': '%s' % uuid.uuid4(),
        'ruleid': '%s' % uuid.uuid4(),
        'src': '127.0.0.1',
        'dst': '127.0.0.1',
        'protocol': 'TCP',
        'port': 9465,
        'connected': True
    }

    def setUp(self):
        self.service = LydianController()
        self.service.start()


    def test_traffic(self):
        import pdb ; pdb.set_trace()
        traffic_rules = [self.DUMMY_RULE]

        with LydianClient('localhost') as client:
            client.controller.register_traffic(traffic_rules)
            time.sleep(2)  # Wait for taffic to run for 10 seconds.
            records = client.results.traffic(self.DUMMY_RULE['reqid'])
            assert records, "Traffic results missing"

    def tearDown(self):
        self.service.stop()



#!/usr/bin/env python

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

version = {}
with open("lydian/__init__.py") as fp:
    exec(fp.read(), version)


setuptools.setup(
    name="lydian",
    version=version['__version__'],
    author="Vipin Sharma, Pradeep Singh, Spiro Kourtessis, Gavin Chang",
    author_email="sharmavipin@vmware.com",
    description="Tool for Traffic Generation, Management and Monitoring",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gitvipin/validation-app-engine",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
)



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.
'''
App for running tcpdump tool and generating .pcap files.
'''
import logging
import os

from lydian.apps.base import exposify
import lydian.apps.console as console


log = logging.getLogger(__name__)


class TCPDumpRerunError(Exception):
    pass

@exposify
class TCPDump(console.Console):

    def __init__(self):
        super(TCPDump, self).__init__()
        self._pcap_handles = {}  # a <str: subprocess.Popen> pair

    def _get_identifier(self, dst_file):
        """
        Maintains identifier scheme for identifying each tcpdump request
        uniquely.
        """
        return os.path.join('/tmp', dst_file)

    def _get_pcap_handle(self, dst_file):
        '''
        Return subprocess.Popen Handle which is running the corresponding
        subprocess for a packet capture request.
        '''
        dst_file = self._get_identifier(dst_file)
        return self._pcap_handles.get(dst_file, None)

    def start_pcap(self, dst_file, interface='eth0', args=''):
        """
        Starts Packet Capture with 'tcpdump' command for given params.
        """
        if self._get_pcap_handle(dst_file):
            msg = "A tcpdump directing to %s is already running" % dst_file
            log.error(msg)
            raise TCPDumpRerunError(msg)

        _dst_file = self._get_identifier(dst_file)
        cmnd = 'tcpdump -i %s %s -w %s' % (interface, args, _dst_file)
        p = self._start_subprocess(cmnd)
        self._pcap_handles[_dst_file] = p
        log.info("Started Packet capture for %s at %s", dst_file, _dst_file)
        return True

    def stop_pcap(self, dst_file):
        """
        Stops packet capture for the destination file `dst_file`.
        """
        ident = self._get_identifier(dst_file)
        proc = self._get_pcap_handle(ident)
        if proc:
            self._ctrl_c(proc)
            self._kill_subprocess(proc)
            self._pcap_handles.pop(ident)
            msg = ("Stopped Packet capture for %s. Result is available"
                   " at %s." % (dst_file, ident))
            log.info(msg)
        else:
            msg = ("No PCAP handle found for %s" % dst_file)
            log.warn(msg)
        return msg

    def is_running(self, dst_file):
        ident = self._get_identifier(dst_file)
        proc = self._get_pcap_handle(ident)
        return self._is_alive(proc)

    def stop(self):
        """
        Stops PCAP app.
        """
        for ident, proc in self._pcap_handles.items():
            log.info("Stopping packet capture for %s", ident)
            self._kill_subprocess(proc, close_fds=True)



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.

import fcntl
import ipaddress
import logging
import platform
import socket
import struct

log = logging.getLogger(__name__)


def is_ipv6_address(address):
    try:
        if address == 'localhost':
            _address = '127.0.0.1'
        elif address == 'ip6-localhost':
            _address = '::1'
        else:
            _address = address
        return ipaddress.ip_address(_address).version == 6
    except Exception as err:
        log.exception(err)
        log.warn("Error while trying to interprete %s as ip address.",
                 address)
        return False


def is_py3():
    return platform.python_version().startswith('3')


def get_mgmt_ifname():
    """" Management Interface name for current platform """
    system = platform.system()
    if system == 'Linux':
        return 'eth0'
    elif system == 'Windows':
        return 'eth0'
    elif system == 'Darwin':
        return 'en0'


def get_host_name():
    return socket.gethostname()



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.
import json
import logging
import time

log = logging.getLogger(__name__)


class Target(object):
    POSIX = 'POSIX'
    NAMESPACE = 'NAMESPACE'
    CONTAINER = 'CONTAINER'
    WINVM = 'WINVM'
    HOST_TYPE = None

    def __init__(self, name=None, ip=None):
        self.name = name
        self.ip = ip    # which IP you login to run operations on.
        self._target_type = self.HOST_TYPE

    @property
    def type(self):
        return self._target_type

    def is_namespace(self):
        return self.type == self.NAMESPACE

    def is_vmhost(self):
        """ Returns True is target is a POSIX VM """
        return self.type == self.POSIX

    def is_container(self):
        return self.type == self.CONTAINER


class VMHost(Target):
    # All POSIX VMs shall have similar way of running
    # Traffic Tasks
    HOST_TYPE = Target.POSIX


class NSHost(Target):
    # Traffic needs to be run differently in case of
    # Namespaces
    HOST_TYPE = Target.NAMESPACE


class ContainerHost(Target):
    # Container Host
    HOST_TYPE = Target.CONTAINER


class WinHost(Target):
    # Container Host
    HOST_TYPE = Target.WINVM


class TrafficRule(object):
    ACTIVE = 'ACTIVE'
    INACTIVE = 'INACTIVE'

    SCHEMA = {
        'ruleid': 'text',       # Unique id of rule.
        'reqid': 'text',        # Request id (non unique e.g. in mesh ping)

        'src': 'text',          # source endpoint
        'dst': 'text',          # destination endpoint
        'protocol': 'text',     # Protocol : TCP/UDP/HTTP
        'port': 'text',         # port
        'connected': 'text',    # Ping expected to PASS or FAIL

        'payload': 'text',      # Payload for rule or "Dinkirk!!"
        'packet': 'text',       # Packet size for traffic.
        'tries': 'text',        # Number of tries/count of ping.

        'username': 'text',     # run traffic as. 'root' by default
        'state': 'text',        # ENABLED/DISABLED

        # Following is more of metadata. Which can be saved into
        # a separate table.
        'src_host': 'text',     # Source host Management IP
        'dst_host': 'text',     # Destination host Management IP

        'purpose':  'text',     # context : CLIENT/SERVER/PERSIST
        'target': 'text',       # Host / Namespace / Container
        }

    DEFAULTS = {
        'state': 'ACTIVE',
        'username': 'root',
        'payload': 'Dinkirk'
    }

    def fill(self):
        """ Sets the default fields if not present."""
        for key, _ in self.SCHEMA.items():
            setattr(self, key, getattr(self, key, None))

        if not self.state:
            self.state = 'ACTIVE'

        if not self.username:
            self.username = 'root'

        # At 255K rules, setting this will need 14 MB of extra memory.
        # likewise, in database, we can optimize for it.
        # if not self.payload:
        #    self.payload = 'Dinkirk!!'

    # Helper methods - Protocol(s)
    def is_TCP(self):
        return self.protocol == 'TCP'

    def is_UDP(self):
        return self.protocol == 'UDP'

    def is_HTTP(self):
        return self.protocol == 'HTTP'

    @property
    def enabled(self):
        return self.state == 'ACTIVE'

    def as_dict(self):
        return {k: v for k, v in self.__dict__.items() if k in self.SCHEMA}

    def __repr__(self):
        return '%s' % self.as_dict()


class Record(object):

    def __init__(self):
        """
        Represents a Base class for recording time series data.
        """
        self._id = None
        self._timestamp = int(time.time())

    def __repr__(self):
        return '%s' % self.toJSON()

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)
    @property
    def timestamp(self):
        return self._timestamp

    def as_dict(self):
        return self.__dict__


class TrafficRecord(Record):

    def __init__(self):
        super(TrafficRecord, self).__init__()
        self.source = None          # ping source IP Address
        self.destination = None     # ping destination IP Address
        self.protocol = None        # ping traffic type (TCP/UDP/HTTP)
        self.port = None            # port on server where Traffic is sent.
        self.expected = None        # ping expectation (Pass/Fail : True/False)
        self.result = None          # Actual result


#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.

import argparse
import logging
import os
import tempfile

from lydian.utils.ssh_host import Host
from lydian.utils.pack import generate_egg

log = logging.getLogger(__name__)


def prep_node(hostip, username='root', password='FAKE_PASSWORD'):

    with Host(host=hostip, user=username, passwd=password) as host:
        try:
            host.req_call('systemctl stop lydian')
        except ValueError:
            pass    # preparing for first time.

        data_dir = os.path.dirname(os.path.realpath(__file__))
        # Copy service file
        service_file = os.path.join(data_dir, '../data/lydian.service')

        host.put_file(service_file,
                      '/etc/systemd/system/lydian.service')

        # Copy Egg file
        egg_file = os.path.join(data_dir, '../data/lydian.egg')
        host.put_file(egg_file, '/tmp/lydian.egg')

        # Copy Config File.
        config_file = os.path.join(data_dir, '../data/lydian.conf')
        host.req_call('mkdir -p /etc/lydian')
        host.put_file(config_file, '/etc/lydian/lydian.conf')

        host.req_call('mv /tmp/lydian.egg ~/')
        try:
            host.req_call('sudo systemctl enable lydian.service')
            host.req_call('sudo systemctl daemon-reload')
            host.req_call('systemctl start lydian')
        except Exception as err:
            log.error("Error in starting service at %s : %r", hostip, err)
            return False
    return True


def cleanup_node(hostip, username='root', password='FAKE_PASSWORD'):
    """
    Cleans up Lydian service from the endpoints.
    """

    with Host(host=hostip, user=username, passwd=password) as host:

        def _func(cmnd):
            try:
                host.req_call(cmnd)
            except ValueError:
                pass
        _func('systemctl stop lydian')
        _func('sudo systemctl disable lydian.service')
        _func('rm /etc/lydian/lydian.conf')
        _func('rm /etc/systemd/system/lydian.service')
        _func('rm /var/log/lydian/lydian.log')


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        '-i', '--hostip', action="store_true", help="IP of host to be prepared.")
    parser.add_argument(
        '-u', '--username', action="store_true", help="Username of host.")
    parser.add_argument(
        '-p', '--password', taction="store_true", help='Password of host.')

    args = parser.parse_args()
    prep_node(args.hostip, args.username, args.password)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.
import logging
import socket
import time

from lydian.traffic.connection import Connection
from lydian.utils.common import is_ipv6_address, is_py3

log = logging.getLogger(__name__)


class PingValidationError(Exception):
    pass


class Client(Connection):
    PAYLOAD = 'Dunkirk!!'
    PING_INTERVAL = 2   # default request rate is 30ppm
    CONNECTION_TIMEOUT = 2

    def __init__(self, server, port, verbose=False, handler=None,
                 interval=None, ipv6=None):
        """
        A simple TCP client which binds to a specified host and port.
        """
        self.server = server
        self.port = port
        try:
            self.interval = int(interval)
        except (ValueError, TypeError) as err:
            _ = err
            self.interval = self.PING_INTERVAL
        self._handler = handler or self.echo_validator
        self._ipv6 = ipv6 or is_ipv6_address(self.server)
        super(Client, self).__init__(verbose=verbose)

    def echo_validator(self, payload, data):
        """
        Ping Validator
        """
        try:
            assert(data == self.PAYLOAD)
            log.info("Sent : %s Received : %s", self.PAYLOAD, data)
        except AssertionError as err:
            _ = err
            raise PingValidationError()

    def start(self, payload=None, tries=None):
        payload = self.PAYLOAD if payload is None else payload
        try:
            tries = int(tries) if tries is not None else None
        except Exception as err:
            _ = err
            tries = None

        while not self.is_event_set():
            if tries is not None:
                if not tries:
                    break
                else:
                    tries -= 1
            self.ping(payload)
            if self.interval:
                time.sleep(self.interval)

    def ping(self, payload):
        raise NotImplementedError("Ping not implemented in %s" % self.__class__.__name__)

    def _prepare_payload(self, payload):
        return payload.encode('utf-8') if is_py3() else payload


class TCPClient(Client):

    def _create_socket(self):
        """
        Returns a simple TCP server socket.
        """
        sock_type = socket.AF_INET6 if self._ipv6 else socket.AF_INET
        self.socket = socket.socket(sock_type, socket.SOCK_STREAM)
        self.socket.settimeout(self.CONNECTION_TIMEOUT)

    def ping(self, payload):
        data = None
        try:
            # create socket
            self._create_socket()
            self.socket.connect((self.server, self.port))
            self.socket.settimeout(None)
            payload = self._prepare_payload(payload)
            self.socket.send(payload)
            data = self.socket.recv(self.MAX_PAYLOAD_SIZE)
            # close socket connection
            self.socket.close()

            if self.verbose:
                msg = "ping to %s:%s pass. data - %r" % (
                    self.server, self.port, data)
                # self.log.info(msg)
            return data  # TODO : is is needed ?
        except Exception as err:
            msg = "ping to %s:%s failed. Error - %r" % (
                self.server, self.port, err)
            if self.verbose:
                self.log.error(msg)
        finally:
            self._handler(payload, data)


class UDPClient(Client):

    def _create_socket(self):
        """
        Returns a simple TCP server socket.
        """
        sock_type = socket.AF_INET6 if self._ipv6 else socket.AF_INET
        self.socket = socket.socket(sock_type, socket.SOCK_DGRAM)
        self.socket.settimeout(self.CONNECTION_TIMEOUT)

    def ping(self, payload):
        try:
            # create socket
            self._create_socket()
            addr = (self.server, self.port)
            payload = self._prepare_payload(payload)
            self.socket.sendto(payload, addr)
            try:
                data, server = self.socket.recvfrom(self.MAX_PAYLOAD_SIZE)
            except Exception:
                data, server = None, None
            _ = server
            # close socket connection
            self.socket.close()

            if self.verbose:
                msg = "ping to %s:%s pass. data - %r" % (
                    self.server, self.port, data)
                self.log.info(msg)
            return data
        except Exception as err:
            self.log.error("ping to %s:%s failed. Error - %r",
                           self.server, self.port, err)
            # TODO : check if raise is an issue.
            # return False
            raise
        finally:
            self._handler(payload, data)



#!/usr/bin/env python
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.
"""
This module defines Console App in AXON system.

This app shall allow running the commands on the endpoints locally. Other apps
which need to work on console such as Running traffic through other tools
(Iperf etc) or running port monioring tools shall also use/inherit this common
app.

NOTE: This should not be exposed via controller as it can pause security risk
as remote endpoints can run malicious commands using the exposed interface.
"""

import logging
import os
import shlex
import signal
import subprocess
import time

from lydian.apps.base import BaseApp

log = logging.getLogger(__name__)


class Console(BaseApp):
    NAME = "CONSOLE"

    def __init__(self):
        """
        Console App runs commands on the console of this node.
        """
        pass

    def run_command(self, cmnd, env=None, cwd=None, timeout=-1):
        """
        This function runs the command "cmnd" and returns the return code and
        result for the command. The command is taken as single string.

        Parameters
        ----------
        cmnd : string
            Command to be executed in string format.
        env : string
            Any environment configuration to be set for environment.
        cwd : string
            Absolute path to the directory where command should be run.
        timeout : int
            time in seconds after which command should be killed. Default of -1
            means command is run without a time limit.

        Returns
        -------
        (int, string)
            returns a tuple of command execution return code and output.
        """

        p = self._start_subprocess(cmnd=cmnd, cwd=cwd, env=env)

        # Start the counter for command to finish if requested.
        if timeout > 0:
            time_limit = time.time() + timeout

            while time.time() < time_limit:
                if self._is_alive(p):
                    time.sleep(1)
                else:
                    break
            self._kill_subprocess(p)

        stdout_val = p.communicate()[0]
        return p.returncode, stdout_val.strip().decode('utf-8')

    def _start_subprocess(self, cmnd,
                          stdin=None, stdout=None, stderr=None,
                          env=None, cwd=None):
        """
        Starts a subprocess and returns a handle for it.
        """
        stdout = subprocess.PIPE if not stdout else stdout
        stderr = subprocess.STDOUT if not stderr else stderr

        cmnd = shlex.split(cmnd)

        return subprocess.Popen(cmnd, shell=False, cwd=cwd, env=env,
                                stdin=stdin, stdout=stdout, stderr=stderr,
                                close_fds=True, preexec_fn=os.setsid,
                                bufsize=-1)

    def _ctrl_c(self, proc, timeout=1):
        """
        """
        proc.send_signal(signal.SIGINT)
        # For child processes, you need to wait for some time after sending
        # SIGINT signal : https://bugs.python.org/issue25942
        proc.wait(timeout)

    def _kill_subprocess(self, proc, close_fds=False):
        """
        Kills a subprocess represented by handle 'proc'.
        """
        # command is still active, kill it.
        log.warning("Killing subprocess %s" % proc.pid)
        proc.kill()
        if close_fds:
            proc.communicate()

    def _is_alive(self, proc):
        """
        Returns True if subprocess is still running else False
        """
        return proc.poll() is None



#!/usr/bin/env python
# Copyright (c) 2019 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2 License
# The full license information can be found in LICENSE.txt
# in the root directory of this project.

from collections import defaultdict
import multiprocessing
import platform
if "Linux" in platform.uname():  # noqa
    from lydian.utils import nsenter
import psutil
import subprocess

INTERFACE_FAMILY = (2, 10)
NAMESPACE_INTERFACE_NAME_PREFIXES = ["veth", "eth"]


def get_interfaces_in_namespace(return_dict):
    return_dict['result'] = psutil.net_if_addrs()
    return return_dict


class Namespace(object):
    """
    Namespace object which holds information name, id, interfaces inside
    """
    def __init__(self, name, id=None):
        self._name = name
        self._id = id
        self._interface_list = []
        self._discover_interfaces()

    @property
    def name(self):
        return self._name

    @property
    def id(self):
        return self._id

    @property
    def interfaces(self):
        return self._interface_list

    def as_dict(self):
        return dict(list(zip(['name', 'id', 'interfaces'],
                        [self.name, self.id, [interface.as_dict() for
                                              interface in self.interfaces]])))

    def _discover_interfaces(self):
        ns_path = '/var/run/netns/%s' % self.name
        manager = multiprocessing.Manager()
        return_dict = manager.dict()
        with nsenter.namespace(ns_path, 'net'):
            process = multiprocessing.Process(
                target=get_interfaces_in_namespace, args=(return_dict,))
            process.start()
            process.join()
        for name, snics in list(return_dict['result'].items()):
            for nic in [snic for snic in snics if snic.family in INTERFACE_FAMILY]:
                self._interface_list.append(Interface(
                    name, nic.address, nic.family,
                    nic.netmask, nic.broadcast))


class NamespaceManager(object):
    """
    Class that controls all relevant NS on a machine
    - upon launch, discovers all namespaces
    - provide apis to get all the necessary information
    """

    def __init__(self):
        self._namespace_map = {}
        self._namespace_interface_map = {}
        self._linux_distro = "Linux" in platform.uname()
        self._discover_namespaces()

    def _discover_namespaces(self):
        if self._linux_distro:
            ns_list = subprocess.check_output(["ip", "netns", "list"])
            if not ns_list:
                return
            for ns in ns_list.decode().rstrip().split("\n"):
                ns_info = str(ns).split()
                id = None
                if len(ns_info) > 1:
                    id = ns_info[1]
                name = ns_info[0]
                self._namespace_map[name] = Namespace(name, id)
                interfaces = Namespace(name, id).interfaces
                if interfaces:
                    self._namespace_interface_map[name] = \
                        Namespace(name, id).interfaces
                else:
                    self._namespace_interface_map[name] = None

    def discover_namespaces(self):
        self._discover_namespaces()

    def get_namespace_interface_map(self):
        return self._namespace_interface_map

    def get_all_namespaces(self):
        """
        Get the list of all namespaces presnt in the system
        :return: List of namespace names
        :rtype: list
        """
        return list(self._namespace_map.keys())

    def get_namespace(self, namespace):
        ns = self._namespace_map.get(namespace)
        if ns:
            return ns.as_dict()
        else:
            return None

    def get_all_namespaces_ips(self):
        """
        Get the list of all namespaces ips address present in the system
        based on NAMESPACE_INTERFACE_NAME_PREFIXES
        :return: List of namespace ips
        :rtype: list
        """
        namespaces_ips = []
        for np in list(self._namespace_map.values()):
            interfaces = [iface for iface in np.interfaces for prefix in
                          NAMESPACE_INTERFACE_NAME_PREFIXES
                          if prefix in iface.name]
            if not interfaces:
                continue
            for iface in interfaces:
                namespaces_ips.append(iface.address)

        return namespaces_ips


class Interface(object):
    """
    Interface object which holds all the information related to interface
    """
    def __init__(self, name, address, family, netmask, broadcast):
        self._name = name
        self._address = address
        self._family = family
        self._netmask = netmask
        self._broadcast = broadcast

    @property
    def name(self):
        return self._name

    @property
    def address(self):
        return self._address

    @property
    def family(self):
        return self._family

    @property
    def netmask(self):
        return self._netmask

    @property
    def broadcast(self):
        return self._broadcast

    def as_dict(self):
        return dict(list(zip(['name', 'address', 'family', 'netmask', 'broadcast'],
                        [self.name, self.address, self.family, self.netmask,
                         self.broadcast])))


class InterfaceManager(object):
    """
    Class that controls all Interface information on host
    """
    def __init__(self):
        self._discover_interfaces()

    def _discover_interfaces(self):
        self._interface_map = defaultdict(list)
        addrs = psutil.net_if_addrs()
        for name, snics in list(addrs.items()):
            for nic in [snic for snic in snics if snic.family in INTERFACE_FAMILY]:
                self._interface_map[name].append(
                    Interface(name, nic.address, nic.family, nic.netmask,
                              nic.broadcast))

    def get_all_interfaces(self):
        """
        Get all the interface names present in the system
        :return:
        :rtype:
        """
        return list(self._interface_map.keys())

    def get_interface(self, name, ip=None):
        """
        Get detail about a particular interface
        :param name: name of the interface
        :type name: str
        :param ip: IPv4 or IPv6 address to be matched
        :type ip: str
        :return: interface object as dict
        :rtype: dict
        """
        snics = self._interface_map.get(name)
        if not snics:
            return None
        for snic in snics:
            if not ip:
                # Return the first interface found in snics if IP address is not
                # provided
                return snic.as_dict()
            else:
                if snic.address == ip:
                    return snic.as_dict()

    def get_interface_by_ip(self, ip):
        for name, interface in list(self._interface_map.items()):
            for sub_interface in interface:
                if sub_interface.address == ip:
                    return name

    def get_ips_by_interface(self, interface):
        ips = []
        if interface not in self._interface_map:
            return ips
        for sub_interface in self._interface_map.get(interface):
            ips.append(sub_interface.address)
        return ips

    def get_all_ips(self):
        all_ips = []
        all_interfaces = self.get_all_interfaces()
        for interface in all_interfaces:
            all_ips.extend(self.get_ips_by_interface(interface))
        return all_ips
